#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Zero-Noise Extrapolation (ZNE) is one of the most powerful techniques for mitigating quantum gate error
in quantum computing. Notice that ZNE does not directly reduce the inherent noise in the quantum
computing process, but rather helps us to obtain the expected computation result by repeating
the same quantum computing process various times with different level of noise. The
advantage of ZNE is that we do not need to know the exact form of the noise as well as how
to control the noise source.

The ZNE method is composed of two steps: rescaling noise and extrapolating.
Among many of noise rescaling techniques, time-variant rescaling is one of the most robust
and promising. This technique stretches the system Hamiltonian in time according to some
rescaling coefficient in order to obtain an equivalently noise-rescaled final quantum state.
For simplicity, we use the Richardson extrapolation in our Quanlse implementation, which is
a mature numeric algorithm that can eliminate error of any order in principle. We remark
that there are many other extrapolation methods such as polynomial and exponential extrapolation.

For more details, please visit https://quanlse.baidu.com/#/doc/tutorial-ZNE.
"""

from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.QOperation import CircuitLine
import numpy as np

from Quanlse.QRpc import rpcCall

from Quanlse.QPlatform.Utilities import (
    numpyMatrixToDictMatrix,
    circuitLineDump,
)

from typing import List, Tuple


def remoteZNEMitigation(rho: np.ndarray,
                        circuit: List[CircuitLine] = None,
                        A: np.ndarray = None,
                        ham: QHam = None,
                        order: int = 1) -> Tuple[float, List[float], List[float]]:
    r"""
    Use the extrapolation method to mitigate the expectation value:

    :math:`{\rm Tr}[A C_n\cdots C_1 \rho C_1^\dagger \cdots C_n^\dagger]`

    under the assumption that the implementation of :math:`C_i` suffer from noise.
    More specifically, the hamiltonian designed to implement the
    quantum circuit :math:`C_n ... C_1` is not perfect.

    If `order` is not set, it will set to 1 by default.
    The Richardson extrapolation method will be applied.
    The return value will be of the form:

    `mitigatedValue, [coefficients], [noisy values]`

    where the `[coefficients]` and `[noisy values]` are the data points
    used to accomplish the Richardson extrapolation, each of size `order+1`.

    :param rho: the input quantum state
    :param circuit: the quantum circuit in Quantum Leaf syntax
    :param A: the quantum observable
    :param ham: the system Hamiltonian designed to implement the quantum circuit.
                It may be generated by Scheduler or optimizers.
    :param order: the order of the error to be extrapolated.
                Must be a positive integer.
    :return: The error mitigated expectation value and possibly
                a list of coefficients and noisy expectation values.
    """
    args = [
        numpyMatrixToDictMatrix(rho),
        circuitLineDump(circuit),
        numpyMatrixToDictMatrix(A),
        ham.dump(),
        order
    ]
    kwargs = {}

    origin = rpcCall("zneMitigation", args, kwargs)

    return origin["mitigatedValue"], origin["infidelities"], origin["noisyValues"]



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utils functions
"""
import copy
from math import atan2, pi
from typing import List, Union, Iterable, Optional, Callable
import itertools
from itertools import product

import numpy
from numpy import all, array, ndarray, zeros, kron, trace, dot, eye, angle, real, exp, \
    cos, arange, fft, argmax, abs, expand_dims, block, identity, meshgrid, sqrt, arctan, conj, prod

from scipy.optimize import fmin
from scipy import linalg
from scipy.special import laguerre, genlaguerre
from math import factorial

from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.QOperator import QOperator, dagger, destroy
from Quanlse.Utils.Plot import plotPop


def generateOperatorKey(subSysNum: int, operator: Union[QOperator, List[QOperator]]) -> str:
    """
    Generate key for operators.

    :param subSysNum: subsystem's number
    :param operator: the corresponding operator object(s)
    :return: returned key
    """
    subSysList = []
    nameList = []
    if isinstance(operator, list):
        for op in operator:
            subSysList.append(op.onSubSys)
            nameList.append(op.name)
    else:
        subSysList.append(operator.onSubSys)
        nameList.append(operator.name)

    keyStr = ""
    forward = 0
    for idx in range(subSysNum):
        if idx in subSysList:
            keyStr = keyStr + chr(8855) + f"{nameList[forward]}{idx}"
            forward += 1

    return keyStr.strip(chr(8855))


def combineOperatorAndOnSubSys(subSysNum: int, operators: Union[QOperator, List[QOperator]] = None,
                               onSubSys: Union[int, List[int]] = None) -> Union[QOperator, List[QOperator]]:
    """
    Set onSubSys information into the operator.

    :param subSysNum: subsystem's number
    :param operators: the corresponding operator object(s)
    :param onSubSys: subsystem's index(es)
    :return: returned QOperator object(s)

    """
    # Verify if operator and onSubSys have the same size
    if isinstance(operators, list) and not isinstance(onSubSys, list):
        raise ArgumentError("operator is a list, however onSubSys is not a list.")
    if not isinstance(operators, list) and isinstance(onSubSys, list):
        raise ArgumentError("onSubSys is a list, however operator is not a list.")

    if isinstance(operators, list) and isinstance(onSubSys, list):
        if len(operators) != len(onSubSys):
            raise ArgumentError(f"The size of operator ({len(operators)}) != that "
                                      f"of onSubSys ({len(onSubSys)})!")

    # Verify the range and set the onSubSys
    if isinstance(onSubSys, int):
        if onSubSys >= subSysNum:
            raise ArgumentError(f"onSubSys ({onSubSys}) is larger than the "
                                      f"subSysNum {subSysNum}.")
        operators.onSubSys = onSubSys
        operatorForSave = operators
    elif isinstance(onSubSys, list):
        if max(onSubSys) >= subSysNum:
            raise ArgumentError(f"onSubSys ({onSubSys}) is larger than the "
                                      f"subSysNum {subSysNum}.")
        # Sort the operators according to on onSubSys
        sortedIndex = list(array(onSubSys).argsort())
        sortedOperators = [operators[i] for i in sortedIndex]
        sortedOnSubSys = [onSubSys[i] for i in sortedIndex]

        # Set the onSubSys
        for i in range(len(sortedIndex)):
            sortedOperators[i].onSubSys = sortedOnSubSys[i]
        operatorForSave = sortedOperators
    else:
        raise ArgumentError("onSubSys should be an integer or a list!")

    return operatorForSave


def formatOperatorInput(operators: Union[QOperator, Callable, List[QOperator], List[Callable]],
                        onSubSys: Union[int, List[int]], sysLevel: Union[int, List[int]]) -> Union:
    """
    We allow input the Callable QOperator instances or the function.
    """
    _operators = []
    if isinstance(operators, list):
        for opIdx, op in enumerate(operators):
            if isinstance(op, QOperator):
                # Input operator is an QOperator instance
                _operators.append(copy.deepcopy(op))
            else:
                # Input operator is callable
                if isinstance(sysLevel, list):
                    _operators.append(op(sysLevel[onSubSys[opIdx]]))
                else:
                    _operators.append(op(sysLevel))
    else:
        if isinstance(operators, QOperator):
            # Input operator is an QOperator instance
            _operators = copy.deepcopy(operators)
        else:
            # Input operator is callable
            if isinstance(sysLevel, list):
                _operators = operators(sysLevel[onSubSys])
            else:
                _operators = operators(sysLevel)
    return _operators


def project(matrix: ndarray, qubitNum: int, sysLevel: Union[int, List[int]], toLevel: int) -> ndarray:
    """
    Project a :math:`d`-level (:math:`d` is an integer) multi-qubit matrix to a lower dimension.

    :param matrix: uReal in ``sysLevel``-dimensional Hilbert space
    :param qubitNum: number of qubits
    :param sysLevel: the energy level of input matrix
    :param toLevel: the target energy level
    :return: uReal in ``toLevel``-dimensional Hilbert space
    """
    if len(matrix.shape) == 1 or min(matrix.shape) == 1:
        isMatrix = False
    else:
        isMatrix = True
    newMat = None
    if isinstance(sysLevel, int):
        if toLevel >= sysLevel:
            raise ArgumentError("The target level should be less than the current level.")
        # Initialization
        if isMatrix:
            tmpM = zeros((sysLevel, sysLevel), dtype=int)
        else:
            tmpM = zeros((sysLevel,), dtype=int)
        # Construct the single qubit matrix.
        if isMatrix:
            for d1 in range(toLevel):
                for d2 in range(toLevel):
                    tmpM[d1, d2] = 1
        else:
            for d1 in range(toLevel):
                tmpM[d1] = 1
        # Construct the tensor product matrix.
        kronMat = array([1], dtype=int) if isMatrix else 1
        for _ in range(qubitNum):
            kronMat = kron(kronMat, tmpM)
        # Output the projected matrix.
        if isMatrix:
            newMat = zeros((toLevel ** qubitNum, toLevel ** qubitNum), dtype=complex)
            toX, toY = 0, 0
            for x in range(sysLevel ** qubitNum):
                dropLine = True
                for y in range(sysLevel ** qubitNum):
                    if kronMat[x, y] == 1:
                        dropLine = False
                        newMat[toX, toY] = matrix[x, y]
                        toY += 1
                toY = 0
                if dropLine is False:
                    toX += 1
        else:
            newMat = zeros((toLevel ** qubitNum,), dtype=complex)
            toX = 0
            for x in range(sysLevel ** qubitNum):
                if kronMat[x] == 1:
                    newMat[toX] = matrix[x]
                    toX += 1
    if isinstance(sysLevel, list):
        if toLevel >= min(sysLevel):
            raise ArgumentError("The target level should be less than the minimum level of one of the qubit.")
        # Construct the tensor product matrix for this system
        kronMat = array([1], dtype=int) if isMatrix else 1
        for level in sysLevel:
            tmpM = zeros((level, level), dtype=int)
            for d1 in range(toLevel):
                for d2 in range(toLevel):
                    tmpM[d1, d2] = 1
            kronMat = kron(kronMat, tmpM)
        # Initialize the output matrix
        newMat = zeros((toLevel ** qubitNum, toLevel ** qubitNum), dtype=complex)
        dim = matrix.shape[0]
        toX, toY = 0, 0
        for x in range(dim):
            dropLine = True
            for y in range(dim):
                if kronMat[x, y] == 1:
                    dropLine = False
                    newMat[toX, toY] = matrix[x, y]
                    toY += 1
            toY = 0
            if dropLine is False:
                toX += 1

    # Return the output matrix
    return newMat


def tensor(*args) -> ndarray:
    """
    Return the tensor product of all matrices in the list.

    :param matrixList: the list of matrices to take the tensor product
    :return: tensor product of matrices in the list
    """
    # We firstly need to check if all the matrix in the list a numpy.ndarray
    if len(args) == 1 and isinstance(args[0], List):
        args = tuple(args[0])
    else:
        if isinstance(args, ndarray):
            return args

    matrixReturn = array([[1.0]], dtype=complex)
    for matrixIndex, matrixItem in enumerate(args):
        matrixReturn = kron(matrixReturn, matrixItem)
    return matrixReturn


def basis(d: int, state: int) -> ndarray:
    r"""
    Return the vector of state operator:
    :math:`|0\rangle, |1\rangle, \cdots`.

    :param d: the dimension of the Hilbert space
    :param state: the index of the state
    :return: matrix form of the state operator
    """
    if d <= state:
        raise ArgumentError("The input state index exceeds system dimension")
    matrix = zeros((d, 1), dtype=complex)
    matrix[state] = 1.0
    return matrix


def projector(a: ndarray, b: ndarray = None) -> ndarray:
    r"""
    Return the matrix form of a state: :math:`|a\rangle \langle b|`.

    :param a: ket operator
    :param b: bra operator
    :return: the outer product of the two operators :math:`|a\rangle` and :math:`(|b\rangle)^{\dagger}`
    """
    if b is None:
        returnMatrix = dot(a, dagger(a))
    else:
        returnMatrix = dot(a, dagger(b))
    return returnMatrix


def expect(matrix: ndarray, state: ndarray) -> float:
    """
    Return the expectation value of the matrix in the given state.

    :param matrix: the given matrix
    :param state: the given state (1-d state or 2-d density matrix)
    :return: expectation value of the matrix in given state
    """

    def isSquare(m: ndarray) -> bool:
        """
        This function is used to check whether the matrix is a square matrix
        :param m: input matrix
        :return: a bool value
        """
        return all(len(row) == len(m) for row in m)

    # We firstly need to check whether the matrix is square
    if not isSquare(matrix):
        raise ArgumentError('Matrix is not a square matrix')
    if state.shape[0] != matrix.shape[1]:
        raise ArgumentError('Dimension Mismatch')

    # is the input state ket or a density matrix?
    if state.shape[0] != state.shape[1]:
        # the input state is a pure state
        state = state @ state.conj().T
        expectValue = real(trace(dot(matrix, state)))

    else:
        # the input state is a density operator
        expectValue = real(trace(dot(matrix, state)))

    return expectValue


def computationalBasisList(qubitNum: int, sysLevel: int) -> List[str]:
    """
    Return a list of strings labeling eigenstates.
    For example, ``computationalBasisList(2, 3)`` will return:
    ``['00', '01', '02', '10', '11', '12', '20', '21', '22']``

    :param qubitNum: the number of qubits in the system
    :param sysLevel: the energy level of the qubits in the system
    :return: the list of strings labeling eigenstates
    """
    if not isinstance(sysLevel, int):
        raise ArgumentError('The system level can only be an integer.')
    itemCount = sysLevel ** qubitNum
    strList = []
    for index in range(itemCount):
        bStr = ''
        for qu in range(qubitNum):
            bStr = f"{int(index / sysLevel ** qu) % sysLevel}{bStr}"
        strList.append(bStr)
    return strList


def generateBasisIndexList(basisStrList: List[str], sysLevel: int) -> List[int]:
    """
    Return a list of integers which indicates the basis indices according to the input basis string list.
    For example, ``generateBasisIndexList(['00', '01', '10', '11'], 3)`` will return:
    ``[0, 1, 3, 4]``

    :param basisStrList: basis string list
    :param sysLevel: the energy level of qubits in the system.
    :return: basis indixes list
    """
    if not isinstance(sysLevel, int):
        raise ArgumentError('The system level can only be an integer.')
        
    strLen = [len(item) for item in basisStrList]
    if max(strLen) != min(strLen):
        raise ArgumentError('All input digital strings should have same length.')
    digLen = max(strLen)

    def translateStrToInt(strN: str) -> int:
        """
        Translate a string to int.
        :param strN: input string
        :return: an int value
        """
        intNum = 0
        for digIndex, charN in enumerate(strN):
            dig = int(charN)
            if dig >= sysLevel:
                raise ArgumentError(f"Digit '{dig}' is greater than sysLevel '{sysLevel}'.")
            intNum += (sysLevel ** (digLen - digIndex - 1)) * dig
        return intNum

    basisIntList = []
    for strNum in basisStrList:
        basisIntList.append(translateStrToInt(strNum))

    return basisIntList


def partialTrace(rho: ndarray, subNum: int, dimList: List[int], index: Union[int, List[int]], mode=0) -> ndarray:
    """
    Partial trace a :math:`d`-level (:math:`d` is an integer) multi-qubit matrix to a lower dimension.

    :param rho: global density matrix to be partially traced
    :param subNum: number of qubits
    :param dimList: the dimension of each subsystem
    :param index: index of subsystem(s) (to be reserved), starts from 1
    :param mode: 0(default)--reserve the subsystems in index list, 1--trace the subsystems in index list
    :return: reserved density matrix
    """
    # the complete indexList
    indexFull = [i + 1 for i in range(subNum)]

    # we use dimListCopy to avoid changing dimList
    dimListCopy = dimList[:]

    a = rho.shape
    if a[0] != a[1]:
        raise ArgumentError("The input matrix is invalid")

    def PT(A, Num, DList, idx):
        c = A.shape

        if Num != len(DList):
            raise ArgumentError("Invalid dimList: dimList is inconsistent with subNum")

        dim = 1
        for k in range(len(DList)):
            if DList[k] == 1 or 0:
                raise ArgumentError("Invalid dimList: Subsystem cannot be one or zero dimension")
            else:
                dim *= DList[k]

        if dim != c[0]:
            raise ArgumentError("Invalid dimList: dimList is inconsistent with rho's dimension")

        if idx <= 0 or idx > Num:
            raise ArgumentError("Invalid index")

        # denote the dimension of the index to be traced as d2
        d2 = DList[idx - 1]
        d1 = 1
        d3 = c[0]

        # denote the dimension in front of the index to be traced as d1
        for k in range(idx - 1):
            d1 = d1 * DList[k]
            d3 = d3 // DList[k]

        # denote the dimension behind the index to be traced as d3
        d3 = d3 // d2

        I1 = eye(d1)
        I2 = eye(d2)

        if d3 != 0:
            I3 = eye(d3)
        else:
            I3 = 1
            d3 = 1
        d = d1 * d3

        # initialize the traced state
        subState = zeros((d, d), dtype=complex)

        # the process of partial trace
        for i in range(d2):
            basis_ = kron(kron(I1, I2[:, i]), I3)
            subState += dot(dot(basis_, A), basis_.T)

        return subState

    if isinstance(index, list):
        index.sort()
        for i in range(len(index)):
            if index[i] == 0:
                raise ArgumentError("Invalid index list: index starts from 1")

        for i in range(len(index)):
            for j in range(len(index)):
                if index[i] == index[j] and i != j:
                    raise ArgumentError("Invalid index list: repeated index")

        if len(index) > subNum:
            raise ArgumentError("Invalid index list: len(index) > subNum")

        if set(index).issubset(indexFull) is False:
            raise ArgumentError("Invalid index list")

        if mode == 0:
            index = list(set(indexFull) - set(index))
            mode = 1

        if mode == 1:
            for i in range(len(index)):
                rho = PT(rho, subNum, dimListCopy, index[i])

                # partition new subsystems, we use dimListCopy to avoid changing dimList from list.pop
                dimListCopy.pop(index[i] - 1)
                subNum = subNum - 1

                # reset the index for next partial trace
                index = array(index) - 1
                index.tolist()
        else:
            raise ArgumentError("Invalid mode")

    elif mode == 1:
        rho = PT(rho, subNum, dimList, index)

    elif mode == 0:
        index = list(set(indexFull) - {index})

        for i in range(len(index)):
            rho = PT(rho, subNum, dimListCopy, index[i])

            # partition new subsystems
            dimListCopy.pop(index[i] - 1)

            subNum = subNum - 1
            # reset the index for next partial trace
            index = array(index) - 1
            index.tolist()

    else:
        raise ArgumentError("Invalid mode")

    return rho


def globalPhase(U: ndarray) -> float:
    r"""
    Compute the global phase of a 2*2 unitary matrix.
    Each 2*2 unitary matrix can be equivalently characterized as:

    :math:`U = e^{i\alpha} R_z(\phi) R_y(\theta) R_z(\lambda)`

    We aim to compute the global phase `\alpha`.
    See also Theorem 4.1 in `Nielsen & Chuang`'s book.

    :param U: the matrix representation of the 2*2 unitary
    :return: the global phase of the unitary matrix
    """
    # Notice that the determinant of the unitary is given by e^{2i\alpha}
    coe = linalg.det(U) ** (-0.5)
    alpha = - angle(coe)
    return alpha


def fromMatrixToAngles(U: ndarray) -> List[float]:
    r"""
    Compute the Euler angles `(\alpha,\theta,\phi,\lambda)` for a single-qubit gate.
    Each single-qubit gate can be equivalently characterized as:

    :math:`U = e^{i\alpha} R_z(\phi) R_y(\theta) R_z(\lambda) \\
    = e^{i(\alpha-\phi/2-\lambda/2)}
    \begin{bmatrix}
    \cos(\theta/2) & -e^{i\lambda}\sin(\theta/2) \\
    e^{i\phi}\sin(\theta/2) & e^{i(\phi+\lambda)}\cos(\theta/2)
    \end{bmatrix}`

    We aim to compute the parameters `(\alpha, \theta,\phi,\lambda)`.
    See Theorem 4.1 in `Nielsen & Chuang`'s book for details.


    :param U: the matrix representation of the qubit unitary
    :return: the Euler angles in List
    """
    if U.shape != (2, 2):
        raise ArgumentError("in fromMatrixToAngles(): input should be a 2x2 matrix!")
    # Remove the global phase
    alpha = globalPhase(U)
    U = U * exp(- 1j * alpha)
    U = U.round(10)
    # Compute theta
    theta = 2 * atan2(abs(U[1, 0]), abs(U[0, 0]))

    # Compute phi and lambda
    phiplambda = 2 * angle(U[1, 1])
    phimlambda = 2 * angle(U[1, 0])
    phi = (phiplambda + phimlambda) / 2.0
    lam = (phiplambda - phimlambda) / 2.0

    return [alpha, theta, phi, lam]


def population(rho: ndarray, subNum: int, dimList: List[int], plot=False) -> dict:
    """
    Output a dictionary to show population of multi-qubit matrix

    :param rho: density matrix
    :param subNum: number of qubits
    :param dimList: the dimension of each subsystem
    :param plot: an option to plot population
    :return: a dictionary illustrate population of each energy level
    """

    a = rho.shape
    if a[0] != a[1]:
        raise ArgumentError("The input matrix is invalid")

    if subNum != len(dimList):
        raise ArgumentError("Invalid dimList: dimList is inconsistent with subNum")

    dim = 1
    for j in range(len(dimList)):
        if dimList[j] == 1 or 0:
            raise ArgumentError("Invalid dimList: Subsystem cannot be one or zero dimension")
        else:
            dim *= dimList[j]

    if dim != a[0]:
        raise ArgumentError("Invalid dimList: dimList is inconsistent with rho's dimension")

    # generate a complete sequence of population
    maxi = max(dimList)
    l_ = zeros(maxi, dtype=int)
    for i in range(maxi):
        l_[i] = '%d' % i
    comp = array(list(product(l_, repeat=subNum)))

    # valuing invalid population
    for i in range(maxi ** subNum):
        for j in range(subNum):
            if comp[i][j] >= dimList[j]:
                comp[i] = 100

    def is_none(n):
        return n[0] == 100

    # filter invalid population, generating a complete sequence of valid population
    pop_name = itertools.filterfalse(is_none, list(comp))
    list_ = list(pop_name)

    popDict = {}

    P = []
    popList = []
    popName = []
    for k in range(dim):
        P.append(basis(dim, k) @ dagger(basis(dim, k)))

    for k in range(dim):
        atr = ''.join(str(i) for i in list_[k])
        pop = abs(trace(rho @ P[k]))
        popList.append(pop)
        popDict[atr] = pop
        popName.append(atr)

    if plot is True:
        # Draw the population of computational basis
        plotPop(popName, popList, xLabel="Computational Basis", yLabel="Population")

    return popDict


def getPopulationOnQubit(popList: List[float], onSubSys: int, sysLevel: int):
    """
    Get the population on the given qubit.

    :param popList: the population list of the full system
    :param onSubSys: the index of subsystem to obtain the population
    :param sysLevel: the level of the system (just supports int at present)
    """

    if isinstance(sysLevel, int):
        targetQLvl = sysLevel
    else:
        raise ArgumentError("sysLevel must be an integer!")

    popRes = [0. for _ in range(targetQLvl)]

    for targetQubitLvlIdx in range(targetQLvl):
        # Traverse all the basis of the target qubit
        for itemIdx, itemVal in enumerate(popList):
            d, m = divmod(int(itemIdx / (sysLevel ** onSubSys)), targetQLvl)
            if m == targetQubitLvlIdx:
                popRes[targetQubitLvlIdx] += itemVal

    return popRes


def findIndex(referenceVecs: ndarray, indexKet: Union[List[ndarray], ndarray]):
    """
    Find the index of the given kets in the indexKet using inner product.

    :param referenceVecs: ndarray of reference vectors.
    :param indexKet: target ket or kets list.
    :return: list of index number.
    """
    # initialize index list
    idxList = []
    refVecList = copy.deepcopy(referenceVecs.T)

    if isinstance(indexKet, list):

        for ket in indexKet:
            innerProd = [vec for vec in refVecList @ ket]
            innerVal = max(innerProd)
            index = innerProd.index(innerVal)
            idxList.append(index)

        return idxList

    else:
        innerProd = [vec for vec in refVecList @ indexKet]
        innerVal = max(innerProd)
        index = innerProd.index(innerVal)
        return index


def fitCR(xData: array, yData: array):
    """
    Fit the rabi oscillation of cross-resonance effect.

    :param xData: The data of x values.
    :param yData: The data of y values.

    :return: The callable fitting function and the fitting parameters.
    """

    def fit(x, p):
        return 0.5 * cos(2 * pi * p[1] * x) + 0.5

    def error(p, x, y, fitFunc):
        return ((y - fitFunc(x, p)) ** 2).sum() / len(y)

    step = xData[1] - xData[0]
    pows = abs(fft.fft(yData))
    freqs = fft.fftfreq(xData.size, step)
    index = argmax(pows[freqs > 0])

    a1Init = abs(yData).max()
    a2Init = freqs[index + 1]

    paraFit = fmin(error, [a1Init, a2Init], (xData, yData, fit))
    return fit, paraFit


def blockDiag(matrix: ndarray, subIndex: List[int]):
    """
    Block diagonalize a given matrix using the principle of least action.

    :param matrix: The given matrix to be block diagonalized.
    :param subIndex: The indexes of sub-system.

    :return: Block diagonalization Matrix and transform Unitary
    """

    if matrix.shape[0] is not matrix.shape[1]:
        raise ArgumentError('Not a square matrix')

    if len(subIndex) > max(matrix.shape):
        raise ArgumentError(f'Number of indexes {len(subIndex)} exceeds matrix dimension len{matrix}')

    # Calculate the eigenvectors and eigenvalues of matrix.
    valsInit, vecsInit = linalg.eig(matrix)

    # Rearrange eigenvalues and corresponding eigenvectors in the ascending order.
    index = valsInit.argsort()
    vecs = vecsInit[:, index]
    vals = valsInit[index]

    s1 = vecs[0:len(subIndex), subIndex]
    s2 = vecs[len(subIndex):len(matrix), subIndex]
    X = -dagger(s2 @ linalg.inv(s1))

    XDag = X.T.conj()
    mat1 = block([identity(X.shape[0]), X])
    mat2 = block([-XDag, identity(XDag.shape[0])])
    U = block([[mat1], [mat2]])
    T = U @ linalg.sqrtm(linalg.inv(dagger(U) @ U))

    # Compute the block-diagonal matrix
    matBD = dagger(T) @ matrix @ T

    return matBD, T


def eigenSystem(matrix: ndarray):
    """
    compute the eigenvalues and the corresponding eigenvectors for the given matrix.
    (sorted by eigenvalues: Ascending order)

    :param matrix: base matrix of the eigen system

    :return: sorted eigenvalues (eigenenergies) and corresponding eigenvectors (eigenstates) (ordered by column)
    """
    # Solve eigenvalues problem
    eigenVals, eigenVecs = linalg.eig(matrix)

    # Rearrange eigenvalues and eigenvectors in the ascending order
    sortedIndex = eigenVals.argsort()
    eigenVecs = eigenVecs[:, sortedIndex]
    eigenVals = eigenVals[sortedIndex]
    eigenVecs = eigenVecs

    return eigenVals.real, eigenVecs


def wigner(rho: ndarray, xRange: ndarray, yRange: ndarray):
    """
    Calculate the wigner function of density matrix using laguerre polynomial.

    :param rho: Input density matrix.
    :param xRange: The range of the X quadrature in the phase space.
    :param yRange: The range of the y quadrature in the phase space.
    """
    X, P = meshgrid(xRange, yRange)
    T = X ** 2 + P ** 2

    def _wignerLaguerre(_m, _n):
        """
        Return wigner function for state |n><m|
        """
        if _m == _n:
            poly = laguerre(_n)
            _w = ((-1) ** _n / pi) * exp(-T) * poly(2 * T)
        else:
            factor1 = sqrt(factorial(m) / factorial(n))
            factor2 = ((-1) ** m / pi) * (sqrt(2) * (X + 1j * P)) ** (n - m) * exp(-T)
            _w = factor1 * factor2 * genlaguerre(m, n - m)(2 * T)
        return _w
    W = zeros([len(xRange), len(yRange)], dtype=complex)
    # Add all the contribution
    dim = prod(rho.shape[0])
    for m in range(dim):
        wmn = real(rho[m, m] * _wignerLaguerre(m, m))
        W += wmn
        for n in range(m + 1, dim):
            wmn = real(rho[m, n] * _wignerLaguerre(m, n))
            W += 2 * wmn
    return W


def coherent(dim: int = 2, alpha: complex = 0 + 0j):
    """
    Generate a coherent state using displacement operator.

    :param dim: dimension of the coherent state truncated.
    :param alpha: the eigenvalue of the annihilation operator.
    """

    # Initialize a vacuum state
    psi0 = basis(dim, 0)
    a = destroy(dim).matrix
    adag = dagger(a)

    # construct a displacement operator
    disp = linalg.expm(alpha * adag - conj(alpha) * a)
    psi = disp @ psi0

    return psi



#!/usr/bin/python3
# -*- coding: utf8 -*-
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Readout resonator Calibration
"""

from Quanlse.Simulator import ReadoutPulse, ReadoutModel
from numpy import ndarray, argmin
from typing import Union, List, Dict, Tuple, Iterable
from math import pi
from scipy.signal import find_peaks
from scipy.optimize import curve_fit


def lorentzian(x: Union[int, float], x0: Union[int, float], a: Union[int, float],
               b: Union[int, float], c: Union[int, float]) -> Union[int, float]:
    """
    The lorentzian function.

    :param x: position.
    :param x0: The center position.
    :param a: The scale of the distribution.
    :param b: The width of the distribution.
    :param c: The shift of the amplitude.

    :return: the y value of the Lorentzian function
    """
    return (a / pi) * (0.5 * b) / ((x - x0) ** 2 + (0.5 * b) ** 2) + c


def fitLorentzian(x: ndarray, y: ndarray) -> Union[ndarray, Iterable, int, float]:
    """
    Fit the curve using Lorentzian function.

    :param x: a list of x data.
    :param y: a list of y data.
    :return: the result of curve fitting.
    """

    yMax = max(y)
    yMaxIdx = find_peaks(y, height=yMax)[0][0]
    yHalf = 0.5 * yMax

    yHalfIdx = argmin(abs(y - yHalf))

    freqCenter = x[yMaxIdx]
    width = 2 * (x[yMaxIdx] - x[yHalfIdx])

    param, cov = curve_fit(lorentzian, x, y, p0=[freqCenter, yMax, width, 0.])

    return param, cov


def resonatorSpec(readoutModel: ReadoutModel, onRes: List[int], freqRange: ndarray,
                  amplitude: Union[int, float], duration: Union[int, float],
                  qubitState='ground', loFreq: Union[int, float] = None) -> Tuple[Dict[int, List], Dict[int, List]]:
    """
    Resonator Spectroscopy.

    :param readoutModel: a ReadoutModel type object.
    :param onRes: Index of the resonators for simulation.
    :param freqRange: drive frequency's range.
    :param amplitude: amplitude of the readout pulse.
    :param duration: duration of the readout pulse.
    :param qubitState: the initial qubit state.
    :param loFreq: lo frequency for demodulation.

    :return: a tuple of vi data and vq data.
    """

    # Initialize readoutPulse object
    driveStrength = {}

    for idx in onRes:
        driveStrength[idx] = amplitude

    if loFreq is not None:
        loFreq = loFreq
    else:
        # The local oscillator frequency for demodulation.
        loFreq = 6.064 * (2 * pi)

    readoutPulse = ReadoutPulse(driveStrength=driveStrength, driveFreq={}, loFreq=loFreq)

    readoutModel.readoutPulse = readoutPulse

    viDict = {}
    vqDict = {}

    for idx in onRes:
        viDict[idx] = []
        vqDict[idx] = []

    for freq in freqRange:

        for idx in onRes:
            readoutModel.readoutPulse.setFreq(idx, freq)

        result = readoutModel.simulate(duration=duration, resIdx=onRes, state=qubitState)

        vi = result['vi']
        vq = result['vq']

        for idx in onRes:
            viDict[idx].append(vi[idx])
            vqDict[idx].append(vq[idx])

    return viDict, vqDict


def findFreq(y: Iterable) -> Tuple[ndarray, dict]:
    """
    Find the index of the peak.

    :param y: a list of signals.
    :return: the index of the peak.
    """

    yMax = max(y)
    yHalf = yMax / 2

    idx = find_peaks(y, height=yHalf)

    return idx



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Global Definitions
"""

import os
import sys

import Quanlse
from Quanlse.QPlatform import Error

env = 'prod'

"""
Environment

Do not modify.

Values: 'prod', 'test'

Used to create or test environment.
"""
if env == "test":
    raise Error.RuntimeError('Not implemented')
else:
    # service address for production
    quantumHubAddr = 'https://quantum-hub.baidu.com/api'
    quantumBucket = 'quantum-task'

sdkVersion = 'Quanlse 2.0.0'
"""
SDK Version

Do not modify.

Used for task submission.
"""

hubToken = os.environ.get('HUBTOKEN', '')
"""
Hub Token

Do not modify.

Used for Quanlse cloud task.

Users can acquire tokens from http://quantum-hub.baidu.com .

Token Management -> Create/View Token
"""

taskSource = os.environ.get('SOURCE', 'PySDK')
taskSourceQuanlse = os.environ.get('SOURCE', 'QuanlseSDK')
"""
Task Source

Do not modify.

Values: 'PySDK', 'PyOnline'

PySDK or PyOnline.
"""

noLocalTask = os.environ.get('NOLOCALTASK', None)
"""
No Local Task

Do not modify.

Values: None or Other

Used for PyOnline.
"""

pollInterval = 5
"""
Poll interval in seconds

Do not modify.

Used for task check.
"""

waitTaskRetrys = 10
"""
Wait task retrys

Do not modify.

Retry count for waiting task in case network failed.
"""

outputPath = os.path.abspath(os.path.join(Quanlse.__file__, '../../Output'))
"""
Output Path

Do not modify by user.

Will be created, when not exist.
"""
if 'sphinx' in sys.modules:
    outputPath = ''
else:
    os.makedirs(outputPath, mode=0o744, exist_ok=True)

circuitPackageFile = os.path.join(outputPath, 'Package.pb')
"""
Circuit Package File

Do not modify.

Circuit hdf5 target file
"""
if 'sphinx' in sys.modules:
    circuitPackageFile = ''


#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
ODE Solver
"""

import time
import copy
from numpy import ndarray, shape, identity, kron, dot, square, round, reshape, array
from typing import Dict, Any

from Quanlse.Utils.Functions import dagger
from Quanlse.Utils.Infidelity import isRho
from Quanlse.QPlatform import Error


def _cacheDictToList(ctrlCache: Dict[str, Any], waveCache: Dict[str, Any]):
    """
    Convert the dict type cache to the list type.
    """
    # Initialize ctrlList and waveList
    _ctrlList, _waveList = [], []
    for ctrlKey in ctrlCache:
        _ctrlList.append(ctrlCache[ctrlKey])
        _waveList.append(array(waveCache[ctrlKey]))
    return _ctrlList, _waveList


def solverNormal(ham: 'QHamiltonian', state0=None, shot=None, recordEvolution=False, accelerate=False):
    """
    Calculate the unitary evolution operator with a given Hamiltonian. This function supports
    both single-job and batch-job processing.

    :param ham: QHamiltonian object
    :param state0: the initial state vector. If None is given, this function will return the time-ordered
                   evolution operator, otherwise returns the final state vector.
    :param recordEvolution: the detailed procedure will be recorded if True
    :param accelerate: use numba expm if true
    :param shot: return the population of the eigenstates when ``shot`` is provided
    :return: result dictionary (or a list of result dictionaries when ``jobList`` is provided)
    """

    # Import expm
    if accelerate:
        from Quanlse.Utils.NumbaSupport import expm
    else:
        from scipy.linalg import expm

    sysLevelLen = 1 if isinstance(ham.sysLevel, int) else len(ham.sysLevel)
    sysLevel = [ham.sysLevel] if isinstance(ham.sysLevel, int) else ham.sysLevel
    subSysNum = ham.subSysNum

    ctrlList, waveList = _cacheDictToList(ham.ctrlCache, ham.waveCache)

    if state0 is not None:
        # Initialize stateOrUnitary
        stateOrUnitary = copy.deepcopy(state0)
    else:
        # If the sysLevel is an integer to indicate the same energy level for each subsystem
        if sysLevelLen == 1:
            stateOrUnitary = identity(sysLevel[0] ** subSysNum, dtype=complex)
        else:
            # If the sysLevel is a list of energy levels for different subsystem
            dim = 1
            for subSysLvl in sysLevel:
                dim = dim * subSysLvl
            stateOrUnitary = identity(dim, dtype=complex)

    # Start time
    startTime = time.time()
    # Create evolution history list
    history = []

    for nowDt in range(0, ham.job.endTimeDt):
        totalHam = ham.driftCache.copy()
        # Traverse all the control Hamiltonian
        for key in range(len(ctrlList)):
            totalHam += ctrlList[key] * waveList[key][nowDt]
        tUnitary = expm(- 1j * totalHam * ham.dt)
        stateOrUnitary = dot(tUnitary, stateOrUnitary)
        if recordEvolution:
            history.append(stateOrUnitary)

    if state0 is not None:
        _state = stateOrUnitary
        _unitary = None
        if shot is not None:
            _population = list(map(int, round(square(abs(_state)) * shot)))
        else:
            _population = None
    else:
        _unitary = stateOrUnitary
        _state = None
        _population = None

    ret = {
        "unitary": _unitary,
        "state": _state,
        "population": _population,
        "dt": ham.dt,
        "dts": ham.job.endTimeDt,
        "ns": ham.job.endTimeDt * ham.dt,
        "time_consuming": time.time() - startTime,
        "evolution_history": history,
        "numba-simulator": "local-built-in-normal"
    }
    return ret


def solverAdaptive(ham: 'QHamiltonian', state0: ndarray = None, shot=None, tolerance: float = 0.01, accelerate=False):
    """
    Run the program calculating the unitary evolution operator for Hamiltonian.
    This is the adaptive algorithm for piecewise-constant, using the pulse sequences given in `hamiltonian`.
    In this algorithm, it applies the strategy of adaptive-step to accelerate the calculation.

    :param ham: QHamiltonian object
    :param state0: the initial state vector. If None is given, this function will return the time-ordered
                   evolution operator, otherwise returns the final state vector.
    :param shot: return the population of the eigenstates when ``shot`` is provided
    :param tolerance: the greatest error for approximation
    :param accelerate: use numba expm if true
    :return: Return a dictionary containing the result.
    """

    # Import expm
    if accelerate:
        try:
            from Quanlse.Utils.NumbaSupport import expm
        except ImportError:
            raise Error.Error("You should install Numba to activate the acceleration; "
                              "Please visit https://numba.pydata.org/ for more information.")
    else:
        from scipy.linalg import expm

    sysLevelLen = 1 if isinstance(ham.sysLevel, int) else len(ham.sysLevel)
    sysLevel = [ham.sysLevel] if isinstance(ham.sysLevel, int) else ham.sysLevel
    subSysNum = ham.subSysNum
    maxDt = ham.job.endTimeDt

    if state0 is not None:
        # Initialize stateOrUnitary
        stateOrUnitary = copy.deepcopy(state0)
    else:
        # If the sysLevel is an integer to indicate the same energy level for each subsystem
        if sysLevelLen == 1:
            stateOrUnitary = identity(sysLevel[0] ** subSysNum, dtype=complex)
        else:
            # If the sysLevel is a list of energy levels for different subsystem
            dim = 1
            for subSysLvl in sysLevel:
                dim = dim * subSysLvl
            stateOrUnitary = identity(dim, dtype=complex)

    # Start time
    startTime = time.time()

    # Start calculation
    calcCount = 0
    nowDt = 0
    while nowDt < maxDt:
        totalHam = ham.driftCache.copy()
        # Decide the step of adaptive algorithm
        maxChange = 0
        forwardStep = 1
        while nowDt < maxDt - forwardStep and maxChange < tolerance:
            for key in ham.ctrlCache:
                currentSeq = ham.waveCache[key][nowDt]
                nextSeq = ham.waveCache[key][nowDt + forwardStep]
                change = abs(nextSeq - currentSeq)
                if change > maxChange:
                    maxChange = change
            forwardStep += 1
        forwardStep -= 1
        forwardStep = max(1, forwardStep)
        # Calculate the total Hamiltonian
        for key in ham.ctrlCache:
            averageSequence = 0
            for innerDt in range(nowDt, nowDt + forwardStep):
                # Obtain the sequence
                currentSequence = ham.waveCache[key][innerDt]
                averageSequence += currentSequence
            totalHam += ham.ctrlCache[key] * averageSequence / forwardStep

        calcCount += 1
        tUnitary = expm(- 1j * totalHam * ham.dt * forwardStep)

        nowDt += forwardStep
        stateOrUnitary = dot(tUnitary, stateOrUnitary)

    if state0 is not None:
        _state = stateOrUnitary
        _unitary = None
        if shot is not None:
            _population = list(map(int, round(square(abs(_state)) * shot)))
        else:
            _population = None
    else:
        _unitary = stateOrUnitary
        _state = None
        _population = None

    ret = {
        "unitary": _unitary,
        "state": _state,
        "population": _population,
        "dt": ham.dt,
        "dts": ham.job.endTimeDt,
        "ns": ham.job.endTimeDt * ham.dt,
        "time_consuming": time.time() - startTime,
        "evolution_history": None,
        "numba-simulator": "local-built-in-adaptive"
    }
    return ret


def solverOpenSystem(ham: 'QHamiltonian', state0=None, recordEvolution=False, accelerate=False):
    """
    Calculate the unitary evolution operator with a given Hamiltonian. This function supports
    both single-job and batch-job processing.

    :param ham: QHamiltonian object
    :param state0: the initial state vector. If None is given, this function will return the time-ordered
                   evolution operator, otherwise returns the final state vector.
    :param recordEvolution: the detailed procedure will be recorded if True
    :param accelerate: use numba expm if true
    :return: result dictionary (or a list of result dictionaries when ``jobList`` is provided)
    """
    if state0.shape[1] == 1:
        rho0 = state0 @ dagger(state0)
    else:
        if isRho(state0):
            rho0 = state0
        else:
            raise Error.ArgumentError('The input state is neither a density matrix nor a state vector')

    # Import expm
    if accelerate:
        from Quanlse.Utils.NumbaSupport import expm
    else:
        from scipy.linalg import expm

    ctrlList, waveList = _cacheDictToList(ham.ctrlCache, ham.waveCache)

    startTime = time.time()
    dt = ham.dt
    qubitsNum = ham.subSysNum

    # Cache information
    ld = ham.dissipationSuperCache
    maxDt = ham.job.endTimeDt
    maxNs = ham.job.endTime
    sysLevel = ham.sysLevel

    # Initialization
    evolutionHistory = []
    propHistotry = []
    dim = None

    # 计算维度
    if isinstance(sysLevel, int):
        dim = sysLevel ** qubitsNum
    elif isinstance(sysLevel, list):
        dim = 1
        for i in sysLevel:
            dim = dim * i

    idn = identity(dim, dtype=complex)
    rhoDim = shape(rho0)[0]

    if rhoDim != dim:
        raise Error.ArgumentError(
            f'The dimension of the input density matrix {rhoDim} is not matched with system level {dim}.')

    _rho = copy.deepcopy(rho0)
    evolutionHistory.append(_rho)
    propt = identity(dim * dim, dtype=complex)

    for nowDt in range(0, maxDt):
        # Initialization
        _ham = ham.driftCache.copy()
        for key in range(len(ctrlList)):
            _ham += ctrlList[key] * waveList[key][nowDt]
        lh = kron(_ham, idn) - kron(idn, _ham.T)
        lind = -1j * lh + ld
        # calculate the propagator
        prop = expm(lind * dt)
        # convert the density matrix to a state vector
        _rhoVec = reshape(_rho, (-1, 1))
        # update the state vector using propagator
        _rhoVec = prop @ _rhoVec
        # convert the state vector back to the density matrix
        _rho = reshape(_rhoVec.T, (dim, dim))
        propt = prop @ propt
        if recordEvolution:
            evolutionHistory.append(_rho)
            propHistotry.append(prop)

    ret = {
        "propagator": propt,
        "state": _rho,
        "dt": ham.dt,
        "dts": maxDt,
        "ns": maxNs,
        "time_consuming": time.time() - startTime,
        "propagator_history": propHistotry,
        "evolution_history": evolutionHistory,
        "numba-simulator": "local-built-in-open-sys"
    }
    return ret



#!/usr/bin/python3
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pulse Generator for QuanlseSchedulerSuperconduct (Randomized Benchmarking).
"""

from Quanlse.Scheduler.Superconduct import GeneratorCloud
from Quanlse.QOperation import CircuitLine
from Quanlse.QWaveform import QJob
from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.Scheduler.SchedulerPulseGenerator import SchedulerPulseGenerator

from Quanlse.remoteOptimizer import remoteOptimize1Qubit as opt1q


def generate1QClifford(ham: QHam = None, cirLine: CircuitLine = None, scheduler: 'SchedulerSuperconduct' = None)\
        -> QJob:
    """
    Default generator for single qubit gates.

    :param ham: QHam object containing the system information
    :param cirLine: a CircuitLine object containing the gate information
    :param scheduler: the instance of Quanlse Scheduler Superconducting
    :return: returned QJob object
    """
    subHam = ham.subSystem(cirLine.qRegIndexList)

    # Use the pulses we generate based on the qubit chosen
    if cirLine.data.name in ['Cinv']:
        job, inf = opt1q(subHam, cirLine.data.getMatrix(), depth=6, targetInfid=0.0002)
    else:
        job, inf = opt1q(ham, cirLine.data.getMatrix(), depth=4, targetInfid=0.0001)

    print(f"Infidelity of {cirLine.data.name} on qubit {cirLine.qRegIndexList}: {inf}")

    subHam.job = job
    return subHam.outputInverseJob(ham.subSysNum, ham.sysLevel, ham.dt)


def SingleQubitCliffordPulseGenerator(ham: QHam) -> SchedulerPulseGenerator:
    """
    Return a single-qubit Clifford pulse SchedulerPulseGenerator instance for the scheduler.

    :param ham: a Hamiltonian object
    :return: returned generator object
    """
    generator = SchedulerPulseGenerator(ham)
    gateList1q = ['X', 'Y', 'Z', 'H', 'S', 'T', 'RX', 'RY', 'RZ', 'W', 'SQRTW', 'U']
    generator.addGenerator(gateList1q, GeneratorCloud.generate1Q)
    gateList1QCliff = ['C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9', 'C10', 'C11', 'C12', 'C13', 'C14', 'C15',
                       'C16', 'C17', 'C18', 'C19', 'C20', 'C21', 'C22', 'C23', 'C24', 'Cinv']
    generator.addGenerator(gateList1QCliff, generate1QClifford)
    generator.addGenerator(['CR'], GeneratorCloud.generateCr)
    generator.addGenerator(['CZ'], GeneratorCloud.generateCz)
    generator.addGenerator(['ISWAP'], GeneratorCloud.generateISWAP)
    return generator



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fixed Gate Operation
"""
import math
from typing import TYPE_CHECKING, Optional

import numpy
import scipy.linalg

from Quanlse.QOperation import QOperation

if TYPE_CHECKING:
    from Quanlse.QRegPool import QRegStorage


class FixedGateOP(QOperation):
    """
    Fixed gates are set in built-in quantum tool chain.

    Only some solid gates with concrete definitions (without parameters) are set here,

    like Identity, Pauli X, Y, Z, Hadamard, Phase, T, CNOT (CX), etc.
    """

    def __init__(self, gate: str, bits: int, matrix: Optional[numpy.ndarray]) -> None:
        """
        Constructor for FixedGateOP
        """
        super().__init__(gate, bits, matrix)

    def __call__(self, *qRegList: 'QRegStorage') -> None:
        self._op(list(qRegList))


X = FixedGateOP('X', 1,
                numpy.array([[0. + 0.j, 1. + 0.j],
                             [1. + 0.j, 0. + 0.j]])
                )

r"""
Pauli-X operator, also called NOT gate, means flipping the qubit.

For example:

:math:`X|0 \rangle = |1\rangle \quad \text{and} \quad  X|1 \rangle = |0\rangle` 

Matrix form:

:math:`X= \begin{bmatrix} 0 & 1 \\ 1 & 0 \end{bmatrix}`
"""

Y = FixedGateOP('Y', 1,
                numpy.array([[0. + 0.j, 0. - 1.j],
                             [0. + 1.j, 0. + 0.j]])
                )
r"""
Pauli-Y operator, similar to Pauli-X operator

Matrix form:

:math:`Y= \begin{bmatrix}     0 & -i \\     i & 0    \end{bmatrix}`
"""

Z = FixedGateOP('Z', 1,
                numpy.array([[1. + 0.j, 0. + 0.j],
                             [0. + 0.j, -1. + 0.j]])
                )  # Pauli-Z operator
r"""
Pauli-Z operator, means changing a local phase

For example: 

:math:`Z(|0 \rangle +|1 \rangle )= |0 \rangle - |1 \rangle`  

Matrix form:

:math:`Z= \begin{bmatrix}     1 & 0 \\     0 & -1    \end{bmatrix}`
"""

H = FixedGateOP('H', 1,
                1 / numpy.sqrt(2) * numpy.array([[1, 1],
                                                 [1, -1]], dtype=complex)
                )
r"""
Hadamard gate: it's the most important single qubit gate. 

And it can prepare a superposed state via applied on zero state, i.e.,

:math:`H|0 \rangle =\frac{1}{\sqrt{2}}( |0 \rangle + |1 \rangle)`  

Matrix form:

:math:`H=\frac{1}{\sqrt{2}} \begin{bmatrix}     1 & 1 \\     1 & -1    \end{bmatrix}`
"""

S = FixedGateOP('S', 1,
                numpy.array([[1. + 0.j, 0. + 0.j],
                             [0. + 0.j, 0. + 1.j]])
                )
r"""
Phase gate or :math:`\frac{\pi}{4}`-gate, it equals :math:`S=e^{-i\frac{\pi}{4}}= Z^{\frac{1}{2}}`.

It changes a local phase, similar to Pauli-Z operator, i.e.,

:math:`S (|0 \rangle +|1 \rangle )= |0 \rangle +i  |1 \rangle`  

Matrix form:

:math:`S=  \begin{bmatrix} 1 & 0 \\ 0 & i \end{bmatrix}`
"""

T = FixedGateOP('T', 1,
                numpy.array([[1. + 0.j, 0. + 0.j],
                             [0. + 0.j, numpy.exp(1j * numpy.pi / 4)]])
                )
r"""
T gate or :math:`(\frac{\pi}{8})`-gate, it equals :math:`T =e^{-i\frac{\pi}{8}}=Z^{\frac{1}{4}}`.

It changes a local phase, similar to :math:`Z` gate, i.e.,

:math:`T (|0 \rangle +|1 \rangle )= |0 \rangle + e^{i\frac{\pi}{4}} |1 \rangle`

Matrix form:

:math:`T = \begin{bmatrix} 1 & 0 \\ 0 & e^{i\frac{\pi}{4}} \end{bmatrix}`
"""

W = FixedGateOP('W', 1,
                (numpy.array([[0. + 0.j, 1. + 0.j],
                              [1. + 0.j, 0. + 0.j]])
                 +
                 numpy.array([[0. + 0.j, 0. - 1.j],
                              [0. + 1.j, 0. + 0.j]]))
                / math.sqrt(2)
                )
r"""
Matrix form:

:math:`W = X + Y = \begin{bmatrix} 0 & \frac{1-i}{\sqrt{2}} \\ \frac{1+i}{\sqrt{2}} & 0 \end{bmatrix}`
"""

SQRTW = FixedGateOP('SQRTW', 1,
                    scipy.linalg.sqrtm(
                        (numpy.array([[0. + 0.j, 1. + 0.j],
                                      [1. + 0.j, 0. + 0.j]])
                         +
                         numpy.array([[0. + 0.j, 0. - 1.j],
                                      [0. + 1.j, 0. + 0.j]]))
                        / math.sqrt(2))
                    )
r"""
Matrix form:

:math:`\sqrt{W} = \sqrt{X + Y} = \sqrt{\begin{bmatrix} 0 & \frac{1-i}{\sqrt{2}} \\ \frac{1+i}{\sqrt{2}} & 0 \end{bmatrix}}`
"""

CZ = FixedGateOP('CZ', 2,
                 numpy.array([
                     [1, 0, 0, 0],
                     [0, 1, 0, 0],
                     [0, 0, 1, 0],
                     [0, 0, 0, -1],
                 ], dtype=complex)
                 )
r"""
CZ gate, or control-Z gate, a native gate in the superconducting platform. It's similar to CNOT gate.

On the simulator mode (use inside the simulator): Matrix form:

:math:`CZ = \begin{bmatrix} 1 & 0  &0 & 0 \\ 0 & 1  & 0& 0 \\ 0 & 0  & 1& 0 \\ 0 & 0  & 0& -1 \end{bmatrix}`
"""

CR = FixedGateOP('CR', 2,
                 numpy.array([
                     [math.cos(-math.pi / 4), -(1j * math.sin(-math.pi / 4)), 0, 0],
                     [-(1j * math.sin(-math.pi / 4)), math.cos(-math.pi / 4), 0, 0],
                     [0, 0, math.cos(-math.pi / 4), (1j * math.sin(-math.pi / 4))],
                     [0, 0, (1j * math.sin(-math.pi / 4)), math.cos(-math.pi / 4)]
                 ], dtype=complex)
                 )
r"""
Cross-resonance gate, a native gate in the superconducting platform.

On the simulator mode (use inside the simulator): Matrix form:

:math:`CR = \begin{bmatrix} \cos(-\pi/4) & -i\sin(-\pi/4)  & 0 & 0 \\ -i\sin(-\pi/4) & \cos(-\pi/4)  & 0 & 0 \\ 
0 & 0  & \cos(-\pi/4) & i\sin(-\pi/4) \\ 0 & 0  & i\sin(-\pi/4) & \cos(-\pi/4) \end{bmatrix}`
"""

EchoedCR = FixedGateOP('EchoedCR', 2,
                 numpy.array([
                     [math.cos(-math.pi / 4), -(1j * math.sin(-math.pi / 4)), 0, 0],
                     [-(1j * math.sin(-math.pi / 4)), math.cos(-math.pi / 4), 0, 0],
                     [0, 0, math.cos(-math.pi / 4), (1j * math.sin(-math.pi / 4))],
                     [0, 0, (1j * math.sin(-math.pi / 4)), math.cos(-math.pi / 4)]
                 ], dtype=complex)
                 )
r"""
echoed cross-resonance gate, a native gate in the superconducting platform.

On the simulator mode (use inside the simulator): Matrix form:

:math:`CR = \begin{bmatrix} \cos(-\pi/4) & -i\sin(-\pi/4)  & 0 & 0 \\ -i\sin(-\pi/4) & \cos(-\pi/4)  & 0 & 0 \\ 
0 & 0  & \cos(-\pi/4) & i\sin(-\pi/4) \\ 0 & 0  & i\sin(-\pi/4) & \cos(-\pi/4) \end{bmatrix}`
"""

ISWAP = FixedGateOP('ISWAP', 2,
                    numpy.array([
                         [1, 0, 0, 0],
                         [0, 0, -1j, 0],
                         [0, -1j, 0, 0],
                         [0, 0, 0, 1],
                    ], dtype=complex)
                    )
r"""
ISWAP gate, a native gate in the superconducting platform.

On the simulator mode (use inside the simulator): Matrix form:

:math:`CZ = \begin{bmatrix} 1 & 0  &0 & 0 \\ 0 & 0  & -i & 0 \\ 0 & -i & 0 & 0 \\ 0 & 0  & 0 & 1 \end{bmatrix}`
"""

CNOT = FixedGateOP('CNOT', 2, numpy.array([
                         [1, 0, 0, 0],
                         [0, 1, 0, 0],
                         [0, 0, 0, 1],
                         [0, 0, 1, 0],
                    ], dtype=complex)
                    )
r"""
CNOT gate, or control-X gate.

On the simulator mode (use inside the simulator): Matrix form:

:math:`CNOT = \begin{bmatrix} 1 & 0  &0 & 0 \\ 0 & 1  & 0& 0 \\ 0 & 0  & 0 & 1 \\ 0 & 0  & 1 & 0 \end{bmatrix}`
"""

SWAP = FixedGateOP('SWAP', 2, numpy.array([                       
                            [1, 0, 0, 0],
                            [0, 0, 1, 0],
                            [0, 1, 0, 0],
                            [0, 0, 0, 1]
                        ], dtype=complex)
                        )
r"""
SWAP gate.

On the simulator mode (use inside the simulator): Matrix form:

:math:`SWAP = \begin{bmatrix} 1 & 0  &0 & 0 \\ 0 & 0 & 1 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \end{bmatrix}`
"""

MS = FixedGateOP('MS', 2,
                 numpy.array([
                     [1, 0, 0, 1.j],
                     [0, 1, 1.j, 0],
                     [0, 1.j, 1, 0],
                     [1.j, 0, 0, 1],
                 ], dtype=complex) / math.sqrt(2)
                 )
r"""
MS gate for Ion.
"""



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Qubit-Coupler-Qubit architecture template and analysis toolkit
"""

import copy
from Quanlse.QPlatform.Error import ArgumentError
from typing import List, Union, Dict, Optional
from numpy import pi, ndarray, identity, trace, diag, array
from Quanlse.Simulator import PulseModel
from Quanlse.QHamiltonian import QHamiltonian
from Quanlse.QPlatform import Error
from Quanlse.QWaveform import gaussian, QJob, quasiSquareErf, virtualZ
from Quanlse.QOperator import create, destroy, dagger, sigmaZ, sigmaX, sigmaY, uWave
from Quanlse.Utils.Functions import tensor, basis, findIndex, blockDiag, eigenSystem
from Quanlse.Scheduler.Superconduct import Scheduler, SchedulerPulseGenerator
from Quanlse.QOperation import CircuitLine
from Quanlse.QOperation.FixedGate import CR, X, EchoedCR
from Quanlse.QOperation.RotationGate import VZ, RX, RY


def pulseGenerator(ham: QHamiltonian) -> SchedulerPulseGenerator:
    r"""
    The pulseGenerator for the simulator.

    :param ham: a QHamiltonian object.
    :return: a SchedulerPulseGenerator object.
    """
    generator = SchedulerPulseGenerator(ham)

    def generateVZ(ham: QHamiltonian, cirLine: CircuitLine) -> QJob:
        """
        Generate the virtual-z gate
        """
        job = ham.createJob()

        onQubit = int(cirLine.qRegIndexList[0])

        phase = cirLine.data.uGateArgumentList[2]

        vz = virtualZ(phase=phase)

        job.appendWave(operators=uWave, onSubSys=onQubit, waves=vz)

        return job

    def generateHadamard(ham: QHamiltonian, cirLine: CircuitLine, scheduler: Scheduler) -> QJob:
        """
        Generate Hadamard gate.
        """

        qIndex = list(cirLine.qRegIndexList)
        jobRX = generate1QGate(ham, CircuitLine(RX(pi), qIndex), scheduler)
        jobRY = generate1QGate(ham, CircuitLine(RY(-pi / 2), qIndex), scheduler)

        return jobRX + jobRY

    def generateCR(ham: QHamiltonian, cirLine, scheduler, ampFactor=1., tFactor=1.) -> QJob:
        """
        Generate the CR gate
        :ham: hamiltonian.
        :cirLine: circuitLine.
        :scheduler: scheduler.
        :ampFactor: the scale of the amplitude.
        :tFactor: the scale of the time gate.
        """

        qIndex = list(cirLine.qRegIndexList)
        if tuple(qIndex) not in scheduler.conf["caliDataCR"].keys():
            qIndex.sort()
        qIndex = tuple(qIndex)

        # Pulse parameters
        CRtg = tFactor * scheduler.conf["caliDataCR"][qIndex]["CRtg"]
        CRamp = ampFactor * scheduler.conf["caliDataCR"][qIndex]["CRamp"]
        CRfreq = scheduler.conf["caliDataCR"][qIndex]["CRfreq"]

        # Add waves
        job = ham.createJob()

        crWave = quasiSquareErf(CRtg, CRamp, 0.1 * CRtg, 0.9 * CRtg, 0.35 * CRamp)
        job.appendWave(uWave, 1, crWave, freq=CRfreq)

        return job

    def generate1QGate(ham: QHamiltonian, cirLine, scheduler) -> QJob:
        """
        Generate the single qubit gate
        """
        gateName = cirLine.data.name
        job = ham.createJob()
        if gateName in ['X', 'RX', 'Y', 'RY']:

            qIndexL = list(cirLine.qRegIndexList)
            qIndex = qIndexL[0]
            if "caliDataXY" not in scheduler.conf.keys():
                raise ArgumentError(f"'caliDataXY' does not exist in scheduler.conf, "
                                    f"hence X/Y controls are not supported.")
            if qIndex not in scheduler.conf["caliDataXY"].keys():
                raise ArgumentError(f"Configuration of qubit(s) {qIndex} does not, "
                                    f"hence X/Y controls are not supported.")
            # Pulse parameters
            piLen = scheduler.conf["caliDataXY"][qIndex]["piLen"]
            piTau = scheduler.conf["caliDataXY"][qIndex]["piTau"]
            piSigma = scheduler.conf["caliDataXY"][qIndex]["piSigma"]

            if gateName in ['X', 'Y']:
                piAmp = scheduler.conf["caliDataXY"][qIndex]["piAmp"]
            elif gateName in ['RX', 'RY']:
                piAmp = cirLine.data.uGateArgumentList[0] / pi * scheduler.conf["caliDataXY"][qIndex]["piAmp"]
            else:
                raise ArgumentError(f"Unsupported gate {gateName}!")

            if gateName in ['X', 'RX']:
                phShift = 0
            elif gateName in ['Y', 'RY']:
                phShift = pi / 2
            else:
                raise ArgumentError(f"Unsupported gate {gateName}!")

            piWave = gaussian(piLen, piAmp, piTau, piSigma)
            job.appendWave(uWave, qIndex, piWave, phase0=phShift)

        elif gateName == 'VZ':

            job = generateVZ(ham, cirLine)

        else:
            raise ArgumentError(f"Unsupported gate {gateName}!")

        return job

    def generateEchoedCR(ham: QHamiltonian, cirLine, scheduler) -> QJob:
        """
        Generate echoed cross-resonance gate
        """
        qIndex = list(cirLine.qRegIndexList)
        jobCR = generateCR(ham=ham, cirLine=CircuitLine(CR, qIndex), scheduler=scheduler, tFactor=0.5)
        jobX = generate1QGate(ham=ham, cirLine=CircuitLine(X, [qIndex[0]]), scheduler=scheduler)
        jobCRMinus = generateCR(ham=ham, cirLine=CircuitLine(CR, qIndex), scheduler=scheduler,
                                ampFactor=-1, tFactor=0.5)

        return jobCR + jobX + jobCRMinus + jobX

    def generateCNOT(ham: QHamiltonian, cirLine, scheduler) -> QJob:
        """
        Generate the CNOT gate using echoed-CR gate and virtual-Z gate.
        """
        qIndex = list(cirLine.qRegIndexList)
        jobVZ = generateVZ(ham=ham, cirLine=CircuitLine(VZ(pi / 2), [qIndex[0]]))
        jobEchoedCR = generateEchoedCR(ham=ham, cirLine=CircuitLine(CR, qIndex), scheduler=scheduler)
        jobRX = generate1QGate(ham=ham, cirLine=CircuitLine(RX(pi / 2), [qIndex[1]]), scheduler=scheduler)

        return jobVZ + jobEchoedCR + jobRX

    generator.addGenerator(['CR'], generateCR)
    generator.addGenerator(['X', 'Y', 'RX', 'RY', 'VZ'], generate1QGate)
    generator.addGenerator(['EchoedCR'], generateEchoedCR)
    generator.addGenerator(['CNOT'], generateCNOT)
    generator.addGenerator(['H'], generateHadamard)

    return generator


def pulseSimQCQ(dt: float = 0.01, frameMode: str = 'lab') -> PulseModel:
    r"""
    Return a template of QCQ model.

    :param dt: a sampling time period.
    :param frameMode: indicates the frame, ``rot`` indicates the rotating frame,
        ``lab`` indicates the lab frame.

    :return: a QCQ PulseModel object.
    """

    # Parameters setting
    couplerLevel = 2
    qubitLevel = 2

    subSysNum = 3
    subSysLevel = [couplerLevel, qubitLevel, qubitLevel]  # Coupler - Control Qubit - Target Qubit

    freqDict = {
        0: 6.3 * (2 * pi),  # Coupler frequency (MHz)
        1: 5.1 * (2 * pi),  # Control qubit frequency (MHz)
        2: 4.9 * (2 * pi)  # Target qubit frequency (MHz)
    }

    anharmDict = {
        0: 0.0 * (2 * pi),  # Coupler anharmonicity (MHz)
        1: -0.33 * (2 * pi),  # Control qubit anharmonicity (MHz)
        2: -0.33 * (2 * pi)  # Target qubit anharmonicity (MHz)
    }

    couplingMap = {
        (0, 1): 0.098 * (2 * pi),  # Coupling strength of coupler and control qubit (MHz)
        (0, 2): 0.083 * (2 * pi),  # Coupling strength of coupler and target qubit (MHz)
        (1, 2): 0.0025 * (2 * pi)  # Coupling strength of control qubit and target qubit (MHz)
    }

    model = PulseModel(subSysNum=subSysNum, sysLevel=subSysLevel, qubitFreq=freqDict, qubitAnharm=anharmDict,
                       driveFreq=freqDict, couplingMap=couplingMap, frameMode=frameMode, pulseGenerator=pulseGenerator,
                       dt=dt)

    model.savePulse = False

    # Calibrated drive frequency for each qubit

    wd0 = 6.3 * (2 * pi)
    wd1 = 31.986542344407095
    wd2 = 30.768373764443712

    if frameMode == 'lab':
        # Single-qubit gate calibration data
        model.conf["caliDataXY"] = {
            1: {"piAmp": 0.2946938775510204, "piLen": 30, "piTau": 30 / 2, "piSigma": 30 / 7, "dragCoef": 0.0},
            2: {"piAmp": 0.29428571428571426, "piLen": 30, "piTau": 30 / 2, "piSigma": 30 / 7, "dragCoef": 0.0}
        }

        # Calibration data for echo CR gate
        model.conf["caliDataCR"] = {
            (1, 2): {"CRtg": 273, "CRamp": 0.15 * 2 * pi / 2, "CRfreq": 4.8948 * 2 * pi - wd1}
        }

        # Set the local oscillator

        model.ham.job.setLO(uWave, 0, freq=wd0)
        model.ham.job.setLO(uWave, 1, freq=wd1)
        model.ham.job.setLO(uWave, 2, freq=wd2)

    elif frameMode == 'rot':
        raise Error.ArgumentError("Only lab frame is supported!")

    else:
        raise Error.ArgumentError("Only lab frame is supported!")

    return model


def computationalSubspace(obj: Union[PulseModel, QHamiltonian]) -> ndarray:
    r"""
    Return the computational subspace matrix

    :param obj: input PulseModel object or QHamiltonian object

    :return: a 4-dimensions ndarray
    """
    if isinstance(obj, PulseModel):
        ham = obj.createQHamiltonian(frameMode='lab')
    elif isinstance(obj, QHamiltonian):
        ham = obj
    else:
        raise Error.ArgumentError("Pleases input a PulseModel object or a QHamiltonian object.")

    if ham.sysLevel == [3, 4, 4]:

        ham.buildCache()
        matQCQ = copy.deepcopy(ham.driftCache)

        # Decoupling coupler
        matQC1Q, unitaryC1 = fastBlockDiagonalization(mat=matQCQ, blockDim=2 * (4 ** 2))
        matQC0Q, unitaryC0 = fastBlockDiagonalization(mat=matQC1Q, blockDim=(4 ** 2))

        # Rearrange
        matQQ = rearrange(matQC0Q[:(4 ** 2), :(4 ** 2)])

        # Obtain the computation subspace matrix
        subMatrix = matQQ[:4, :4]

    else:
        raise Error.ArgumentError("Coupler (index: 0) level should be 3 and Qubit (index: 1, 2) level should be 4.")

    return subMatrix


def staticZZ(obj: Union[PulseModel, QHamiltonian]) -> float:
    r"""
    Compute the static ZZ strength of the given QCQ model

    :param obj: input a PulseModel object or QHamiltonian object

    :return: a float
    """

    if isinstance(obj, PulseModel):
        ham = obj.createQHamiltonian(frameMode='lab')
    elif isinstance(obj, QHamiltonian):
        ham = obj
    else:
        raise Error.ArgumentError("Pleases input a PulseModel object or a QHamiltonian object.")

    if ham.sysLevel == [3, 4, 4]:

        ham.buildCache()
        matQCQ = copy.deepcopy(ham.driftCache)

        # Create Qubit-Qubit subspace basis vectors
        vec00 = tensor(basis(3, 0), basis(4, 0), basis(4, 0))
        vec01 = tensor(basis(3, 0), basis(4, 0), basis(4, 1))
        vec10 = tensor(basis(3, 0), basis(4, 1), basis(4, 0))
        vec11 = tensor(basis(3, 0), basis(4, 1), basis(4, 1))

        # Get the eigen system of QCQ Hamiltonian
        originEigenVals, originEigenVecs = eigenSystem(matQCQ)

        # vector indexes of QQ eigenvectors
        idx00 = findIndex(originEigenVecs, [vec00])[0]
        idx10 = findIndex(originEigenVecs, [vec10])[0]
        idx01 = findIndex(originEigenVecs, [vec01])[0]
        idx11 = findIndex(originEigenVecs, [vec11])[0]

        staticZZCrossTalk = originEigenVals[idx00] - originEigenVals[idx10] - \
                            originEigenVals[idx01] + originEigenVals[idx11]

    else:
        raise Error.ArgumentError("Coupler (index: 0) level should be 3 and Qubit (index: 1, 2) level should be 4.")

    return staticZZCrossTalk


def effectiveCoupling(obj: Union[PulseModel, QHamiltonian]) -> float:
    r"""
    Return effective coupling strength of Qubit-Qubit Model

    :param obj: input PulseModel object or QHamiltonian object

    :return: a float
    """

    matrixXY = (tensor(sigmaX().matrix, sigmaX().matrix) + tensor(sigmaY().matrix, sigmaY().matrix)) / 2
    subMatrix = computationalSubspace(obj)

    effectiveJ = trace(matrixXY @ subMatrix).real / 2

    return effectiveJ


def dressedQubitFreq(obj: Union[PulseModel, QHamiltonian]) -> Dict:
    """
    Return the dressed frequencies of the control qubit and the target qubit

    :param obj: input a PulseModel object or QHamiltonian object

    :return: dressed frequencies dictionary
    """
    if isinstance(obj, PulseModel):
        ham = obj.createQHamiltonian(frameMode='lab')
    elif isinstance(obj, QHamiltonian):
        ham = obj
    else:
        raise Error.ArgumentError("Pleases input a PulseModel object or a QHamiltonian object.")

    if ham.sysLevel == [3, 4, 4]:

        ham.buildCache()
        matQCQ = copy.deepcopy(ham.driftCache)

        # Create Qubit-Qubit subspace basis vectors
        vec00 = tensor(basis(3, 0), basis(4, 0), basis(4, 0))
        vec01 = tensor(basis(3, 0), basis(4, 0), basis(4, 1))
        vec10 = tensor(basis(3, 0), basis(4, 1), basis(4, 0))
        vec11 = tensor(basis(3, 0), basis(4, 1), basis(4, 1))

        # Get the eigen system of QCQ Hamiltonian
        originEigenVals, originEigenVecs = eigenSystem(matQCQ)

        # Vector indexes of QQ eigenvectors
        idx00 = findIndex(originEigenVecs, [vec00])[0]
        idx10 = findIndex(originEigenVecs, [vec10])[0]
        idx01 = findIndex(originEigenVecs, [vec01])[0]
        idx11 = findIndex(originEigenVecs, [vec11])[0]

        dressedCtrlFreq = (originEigenVals[idx10] - originEigenVals[idx00]
                           + originEigenVals[idx11] - originEigenVals[idx01]) / 2

        dressedTargetFreq = (originEigenVals[idx11] - originEigenVals[idx10]
                             + originEigenVals[idx01] - originEigenVals[idx00]) / 2

        dressedFreqDict = {
            "control": dressedCtrlFreq,
            "target": dressedTargetFreq
        }

    else:
        raise Error.ArgumentError("Coupler (index: 0) level should be 3 and Qubit (index: 1, 2) level should be 4.")

    return dressedFreqDict


def pauliCoefficient(obj: Union[PulseModel, QHamiltonian], drivingAmp: Optional[float] = None) -> Dict:
    r"""
    Return Pauli coefficients of the Hamiltonian with CR drive

    :param obj: input PulseModel object or QHamiltonian object
    :param drivingAmp: amplitude of CR pulse

    :return: Pauli coefficients dictionary
    """

    def _diagnoalize(mat: ndarray):
        """
        Diagonalize the given matrix

        :param mat: input square matrix

        :return: diagonal matrix and the corresponding transformation unitary
        """
        matCache = copy.deepcopy(mat)
        eigenVals, eigenVecs = eigenSystem(matCache)
        indexVecList = createBasis(mat.shape[0], mat.shape[0])
        indexList = findIndex(eigenVecs, indexVecList)
        diagMat = diag(eigenVals[indexList])
        unitary = eigenVecs[:, indexList]
        return diagMat, unitary

    # Define pauli operator
    pauliOpZZ = tensor(sigmaZ().matrix, sigmaZ().matrix)
    pauliOpZX = tensor(sigmaZ().matrix, sigmaX().matrix)
    pauliOpZY = tensor(sigmaZ().matrix, sigmaY().matrix)
    pauliOpZI = tensor(sigmaZ().matrix, identity(2))
    pauliOpIZ = tensor(identity(2), sigmaZ().matrix)
    pauliOpIX = tensor(identity(2), sigmaX().matrix)
    pauliOpIY = tensor(identity(2), sigmaY().matrix)

    # Define the cross-resonance driving operator
    driveCRop = tensor((create(4).matrix + destroy(4).matrix),
                       identity(4))

    # Initial Hamiltonian
    if isinstance(obj, PulseModel):
        ham = obj.createQHamiltonian(frameMode='lab')
    elif isinstance(obj, QHamiltonian):
        ham = obj
    else:
        raise Error.ArgumentError("Pleases input a PulseModel object or a QHamiltonian object.")

    if ham.sysLevel == [3, 4, 4]:

        ham.buildCache()
        matQCQ = copy.deepcopy(ham.driftCache)

        # Decoupling coupler
        matQC1Q, unitaryC1 = fastBlockDiagonalization(mat=matQCQ, blockDim=2 * (4 ** 2))
        matQC0Q, unitaryC0 = fastBlockDiagonalization(mat=matQC1Q, blockDim=(4 ** 2))

        # Rearrange
        matQQ = rearrange(copy.deepcopy(matQC0Q[:(4 ** 2), :(4 ** 2)]))
        driveRearrange = rearrange(driveCRop)

        # Diagonalize the Q-Q subspace matrix
        matDiagQQ, unitaryDiagQQ = _diagnoalize(matQQ)

        if drivingAmp is not None:
            # Transform driving term
            driveMat = dagger(unitaryDiagQQ) @ driveRearrange @ unitaryDiagQQ

            # Initialize rotating term
            dressFreq = (matDiagQQ[1, 1] - matDiagQQ[0, 0] + matDiagQQ[3, 3] - matDiagQQ[2, 2]) / 2
            numberOperator = diag(array([0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 6]))

            rotMat = matDiagQQ + drivingAmp * 0.5 * driveMat - dressFreq * numberOperator
            computeBlockMat, computeSubT = computeBlockDiagonalize(rotMat, 4)
            computeMat = computeBlockMat[:4, :4]

            if obj.qubitFreq[1] >= obj.qubitFreq[2]:
                finalMat, T = blockDiag(computeMat, [0, 1])
            else:
                finalMat, T = blockDiag(computeMat, [2, 3])

            pauliDict = {
                'ZI': trace(pauliOpZI @ finalMat).real / 2,
                'IZ': trace(pauliOpIZ @ finalMat).real / 2,
                'ZZ': trace(pauliOpZZ @ finalMat).real / 2,
                'ZX': trace(pauliOpZX @ finalMat).real / 2,
                'IX': trace(pauliOpIX @ finalMat).real / 2,
                'ZY': trace(pauliOpZY @ finalMat).real / 2,
                'IY': trace(pauliOpIY @ finalMat).real / 2
            }
        else:
            finalMat = matDiagQQ[:4, :4]

            pauliDict = {
                'ZI': trace(pauliOpZI @ finalMat).real / 2,
                'IZ': trace(pauliOpIZ @ finalMat).real / 2,
                'ZZ': trace(pauliOpZZ @ finalMat).real / 2,
                'ZX': trace(pauliOpZX @ finalMat).real / 2,
                'IX': trace(pauliOpIX @ finalMat).real / 2,
                'ZY': trace(pauliOpZY @ finalMat).real / 2,
                'IY': trace(pauliOpIY @ finalMat).real / 2
            }
    else:
        raise Error.ArgumentError("Coupler (index: 0) level should be 3 and Qubit (index: 1, 2) level should be 4.")

    return pauliDict


def createBasis(dim: int, count: int) -> List[ndarray]:
    """
    Create basis vectors list

    :param dim: dimension of the Hilbert space
    :param count: the number of the basis vectors

    :return: a ndarray List
    """
    vecList = []
    for num in range(count):
        vec = basis(dim, num)
        vecList.append(vec)
    return vecList


def rearrange(mat: ndarray) -> ndarray:
    """
    Rearrange the Q-Q Hamiltonian matrix by the excitation number

    :param mat: input Hamiltonian matrix

    :return: rearranged matrix
    """
    if mat.shape == (16, 16):
        matCache = copy.deepcopy(mat)

        matCache[:, [2, 4]] = matCache[:, [4, 2]]
        matCache[[2, 4], :] = matCache[[4, 2], :]

        matCache[:, [3, 5]] = matCache[:, [5, 3]]
        matCache[[3, 5], :] = matCache[[5, 3], :]

        matCache[:, [5, 8]] = matCache[:, [8, 5]]
        matCache[[5, 8], :] = matCache[[8, 5], :]

        matCache[:, [7, 12]] = matCache[:, [12, 7]]
        matCache[[7, 12], :] = matCache[[12, 7], :]

        matCache[:, [6, 8]] = matCache[:, [8, 6]]
        matCache[[6, 8], :] = matCache[[8, 6], :]

        matCache[:, [7, 9]] = matCache[:, [9, 7]]
        matCache[[7, 9], :] = matCache[[9, 7], :]

        matCache[:, [7, 8]] = matCache[:, [8, 7]]
        matCache[[7, 8], :] = matCache[[8, 7], :]

        matCache[:, [11, 13]] = matCache[:, [13, 11]]
        matCache[[11, 13], :] = matCache[[13, 11], :]

    else:
        raise Error.ArgumentError("The matrix dimension should be (16, 16)")

    return matCache


def fastBlockDiagonalization(mat: ndarray, blockDim: int) -> [ndarray, ndarray]:
    """
    Block diagonalize the given matrix

    :param mat: original matrix
    :param blockDim: dimension of the first block of the matrix

    :return: the block diagonalized matrix
    """
    matCache = copy.deepcopy(mat)
    eigenVals, eigenVecs = eigenSystem(matCache)
    subVecList = createBasis(mat.shape[0], blockDim)
    indexList = findIndex(eigenVecs, subVecList)
    matBD, transMat = blockDiag(matCache, indexList)

    return matBD, transMat


def computeBlockDiagonalize(mat: ndarray, blockDim: int = 4) -> [ndarray, ndarray]:
    """
    Block diagonalize the qubit-qubit subspace matrix and get the computational subspace block matrix

    :param mat: Qubit-Qubit subspace matrix
    :param blockDim: dimension of the computational subspace matrix (default: 4)

    :return: the block diagonalized matrix and transformation unitary
    """
    matCache = copy.deepcopy(mat)
    eigenVals, eigenVecs = eigenSystem(matCache)
    subVecList = createBasis(mat.shape[0], blockDim)
    indexList = findIndex(eigenVecs, subVecList)

    # Fix findIndex error
    if len(indexList) == 4:
        if indexList[0] == indexList[1]:
            if indexList[0] == mat.shape[0] - 1:
                indexList[0] = mat.shape[0] - 2
            elif indexList[0] == 0:
                indexList[1] = 1
            else:
                leftDiff = abs(eigenVals[indexList[0] - 1] - eigenVals[indexList[0]])
                righrDiff = abs(eigenVals[indexList[0]] - eigenVals[indexList[0] + 1])
                if leftDiff < righrDiff:
                    indexList[0] = indexList[0] - 1
                else:
                    indexList[0] = indexList[0] + 1
        elif indexList[2] == indexList[3]:
            if indexList[2] == mat.shape[0] - 1:
                indexList[2] = mat.shape[0] - 2
            elif indexList[2] == 0:
                indexList[3] = 1
            else:
                leftDiff = abs(eigenVals[indexList[2] - 1] - eigenVals[indexList[2]])
                rightDiff = abs(eigenVals[indexList[2]] - eigenVals[indexList[2] + 1])
                if leftDiff < rightDiff:
                    indexList[2] = indexList[2] - 1
                else:
                    indexList[2] = indexList[2] + 1
    elif len(indexList) == 2:
        if indexList[0] == indexList[1]:
            leftDiff = abs(eigenVals[indexList[0] - 1] - eigenVals[indexList[0]])
            rightDiff = abs(eigenVals[indexList[0]] - eigenVals[indexList[0] + 1])
            if indexList[0] == indexList[1]:
                if indexList[0] == mat.shape[0] - 1:
                    indexList[0] = mat.shape[0] - 2
                elif indexList[0] == 0:
                    indexList[1] = 1
            else:
                if leftDiff < rightDiff:
                    indexList[0] = indexList[0] - 1
                else:
                    indexList[0] = indexList[0] + 1

    matBD, transMat = blockDiag(matCache, indexList)

    return matBD, transMat




#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities
"""

from typing import Dict, List, Optional, Union, Type

import pickle
import base64
import numpy

from Quanlse.QPlatform import Error
from Quanlse.QOperation import CircuitLine


def numpyMatrixToDictMatrix(numpyMatrix: numpy.ndarray) -> Dict:
    """
    Convert numpy matrix to a dictionary. Must be C-contiguous.

    :param numpyMatrix: numpy matrix
    :return: returned dictionary
    """

    if not numpyMatrix.flags["C_CONTIGUOUS"]:
        raise Error.ArgumentError('Matrix must C-contiguous!')

    if numpyMatrix.size == 0:
        return {}

    array = []
    dictMatrix = {
        'shape': list(numpyMatrix.shape),
        'array': array
    }

    for value in numpy.nditer(numpyMatrix):
        val = value.reshape(1, 1)[0][0]
        array.append({
            'real': val.real,
            'imag': val.imag
        })
    return dictMatrix


def dictMatrixToNumpyMatrix(dictMatrix: Dict, valueType: Union[Type[complex], Type[float]]) -> numpy.ndarray:
    """
    Convert numpy matrix to a dictionary. Must be C-contiguous.

    :param dictMatrix: dictionary consisting the matrix
    :param valueType: select value type
    :return: returned numpy matrix
    """

    if len(dictMatrix) == 0:
        return numpy.empty(0, valueType)

    if valueType == complex:
        complexArray = [complex(complexValue['real'], complexValue['imag']) for complexValue in dictMatrix['array']]
    else:
        complexArray = [complexValue['real'] for complexValue in dictMatrix['array']]
    return numpy.array(complexArray).reshape(dictMatrix['shape'])


def circuitLineDump(cirLine: Union[CircuitLine, List[CircuitLine]]) -> str:
    """
    Convert CircuitLine Object (or the list of that) into a string.

    :param cirLine: a CircuitLine Object
    :return: returned string object
    """
    byteStr = pickle.dumps(cirLine)
    base64str = base64.b64encode(byteStr)
    return base64str.decode()


def circuitLineLoad(base64str: str) -> Union[CircuitLine, List[CircuitLine]]:
    """
    Convert base64 string into CircuitLine objects.
    
    :param base64str: input base64 string
    :return: returned CircuitLine object
    """
    byteStr = base64.b64decode(base64str.encode())
    obj = pickle.loads(byteStr)  # type: Union[CircuitLine, List[CircuitLine]]
    return obj


#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simulate the Hamiltonian using the Quanlse remote simulator.
"""

from numpy import ndarray

from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.QWaveform import QJobList, QResult, QJob
from Quanlse.QRpc import rpcCall
from Quanlse.QPlatform.Utilities import numpyMatrixToDictMatrix


def remoteSimulatorRunHamiltonian(ham: QHam, state0: ndarray = None, job: QJob = None,
                                  jobList: QJobList = None, isOpen=False) -> QResult:
    """
    Simulate the Hamiltonian using the Quanlse remote simulator.

    :param ham: the QHamiltonian object.
    :param state0: The initial state.
    :param job: the QJob object.
    :param jobList: The QJobList object.
    :param isOpen: Run the simulation of open system using Lindblad master equation if true.
    :return: the QResult object.
    """

    maxEndTime = None
    kwargs = {}
    if state0 is not None:
        kwargs["state0"] = numpyMatrixToDictMatrix(state0)
    if job is not None:
        maxEndTime, _ = job.computeMaxTime()
        kwargs["job"] = job.dump(maxEndTime=maxEndTime)
    if jobList is not None:
        maxTimeList = []
        for job in jobList.jobs:
            maxTime, _ = job.computeMaxTime()
            maxTimeList.append(maxTime)
        maxEndTime = max(maxTimeList)
        kwargs["jobList"] = jobList.dump(maxEndTime=maxEndTime)
    if maxEndTime is None:
        maxEndTime, _ = ham.job.computeMaxTime()
    kwargs["isOpen"] = isOpen

    args = [ham.dump(maxEndTime=maxEndTime)]
    origin = rpcCall("runHamiltonian", args, kwargs)

    return QResult.load(origin["result"])



#!/usr/bin/python3
# -*- coding: utf8 -*-
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Single-qubit Calibration
"""

import numpy as np
from scipy.signal import find_peaks
from scipy.optimize import curve_fit
from Quanlse.QOperator import driveX, driveZ
from Quanlse.QWaveform import gaussian, square
from Quanlse.Simulator import PulseModel
from Quanlse.Utils.Functions import basis, dagger, expect
from numpy import linspace, array
from math import pi
from typing import Union, List, Any, Optional
from Quanlse.Define import Error
from scipy import optimize


def ampRabi(pulseModel: PulseModel, pulseFreq: Union[int, float], ampRange: list, tg: int,
            sample: int = 100) -> [list, list]:
    """
    Perform a Rabi Oscillation by varying the pulse amplitudes. This function returns a list of amplitudes scanned
    and a list of populations.

    :param pulseModel: a pulseModel object
    :param pulseFreq: frequency of the pulse
    :param ampRange: a list of amplitude bounds
    :param tg: fixed pulse duration
    :param sample: sample size (default at 100)
    :return: a list of amplitudes scanned and a list of populations
    """
    # initialize population list and define amplitude list
    popList = []
    ampList = linspace(ampRange[0], ampRange[1], sample)

    # define job list
    qJobList = pulseModel.ham.createJobList()
    for amp in ampList:
        qJob = pulseModel.ham.createJob()
        qJob.setLO(driveX, 0, freq=pulseFreq)
        qJob.appendWave(driveX, 0, gaussian(tg, amp, tg / 2, tg / 8))
        qJobList.addJob(qJob)
    stateInit = basis(3, 0)

    # run simulation
    results = pulseModel.ham.simulate(jobList=qJobList, state0=stateInit)

    # obtain population from results
    for result in results:
        rho = result['state']
        prj = basis(3, 1) @ dagger(basis(3, 1))
        popList.append(expect(prj, rho))
    return list(ampList), popList


def tRabi(pulseModel: PulseModel, pulseFreq: Union[int, float], tRange: list, amp: float,
          sample: int = 100) -> [list, list]:
    """
    Perform a Rabi Oscillation by varying the pulse duration. This function returns a list of pulse duration scanned
    and a list of populations.

    :param pulseModel: a pulseModel object
    :param pulseFreq: frequency of the pulse
    :param amp: fixed pulse amplitude
    :param tRange: a list of duration bounds
    :param sample: sample size (default at 100)
    :return: a list of pulse duration scanned and a list of populations
    """

    # initialize population list and define time list
    popList = []
    tList = linspace(tRange[0], tRange[1], sample)

    # define job list
    qJobList = pulseModel.ham.createJobList()
    for t in tList:
        qJob = pulseModel.ham.createJob()
        qJob.setLO(driveX, 0, freq=pulseFreq)
        qJob.appendWave(driveX, 0, gaussian(t, amp, t / 2, t / 8))
        qJobList.addJob(qJob)
    stateInit = basis(3, 0)

    # run simulation
    results = pulseModel.ham.simulate(jobList=qJobList, state0=stateInit)

    # obtain population from results
    for result in results:
        rho = result['state']
        prj = basis(3, 1) @ dagger(basis(3, 1))
        popList.append(expect(prj, rho))
    return list(tList), popList


def fitRabi(popList: list, xList: list, guess: Optional[Union[List[Any], np.ndarray]] = None,
            sample: int = 5000, method: str = 'minimize') -> [float, float]:
    r"""
    Find pi-pulse duration/amplitude with Rabi results. This function takes a list of population, a list of whatever
    is scanned, sample size and a list of initial parameters for the fitting function. The fitting function takes
    form: :math:`y = a \cdot \cos(b \cdot x + c) + d`. The function returns the pi/2 and pi pulse amplitude(duration).

    :param popList: a list of population returned from the Rabi function.
    :param xList: a list of scanned variables(durations or amplitudes)
    :param guess: initial guess of the fitting function.
    :param sample: sample size
    :param method: method for optimization, 'minimize' or dual_annealing
    :return: the pi/2 and pi pulse amplitude(duration)
    """

    # Estimate the oscillation frequency from the given data

    def _estimateInit(_x, _y):
        yMax = max(_y)
        idx = find_peaks(_y, yMax / 2)[0][0]
        halfPeriod = _x[idx] - _x[0]
        period = 2 * halfPeriod
        freq = 1 / period
        aEst = -0.5
        bEst = 2 * pi * freq
        cEst = 0.
        dEst = 0.5
        estInit = np.array([aEst, bEst, cEst, dEst])
        return estInit

    # Define the function to be fitted
    def fit(x, a, b, c, d):
        return a * np.cos(b * x + c) + d

    # Define the loss function
    def loss(x):
        estList = np.array([fit(t, x[0], x[1], x[2], x[3]) for t in xList])
        _loss = np.sum(np.abs(estList - popList))
        return _loss

    # Fit the data
    if method == 'dual_annealing':
        opt = optimize.dual_annealing(loss, [(-100., 100.), (-100., 100.), (-100., 100.), (-100., 100.)], maxiter=4000)

    elif method == 'minimize':
        if guess is None:
            guess = _estimateInit(xList, popList)
            opt = optimize.minimize(loss, x0=guess)
        else:
            opt = optimize.minimize(loss, x0=guess)

    else:
        raise Error.ArgumentError('The method is \'dual_annealing\' or \'minimize\'.')

    def yFit(x):
        return fit(x, opt['x'][0], opt['x'][1], opt['x'][2], opt['x'][3])

    sampleList = np.linspace(xList[0], max(xList), sample)
    y = [yFit(x) for x in sampleList]
    peak = find_peaks(y)[0][0]
    x180 = sampleList[peak]
    x90 = x180 / 2

    return x90, x180


def longRelax(pulseModel: PulseModel, dt: float, pulseFreq: float, piAmp: float, piLen: float,
              maxIdle: float, initFit: list) -> [float, list, list, list]:
    r"""
    Find T1 using longitudinal relaxation, note that this function includes the fitting function.
    The fitting function takes form: :math:`y = \exp(-x / T_1 )`.

    :param pulseModel: a pulse model object
    :param dt: AWG sampling time
    :param pulseFreq: pulse frequency
    :param piAmp: pi-pulse amplitude
    :param piLen: pi-pulse length
    :param maxIdle: max idling time after the pi-pulse
    :param initFit: an initial guess of T1
    :return: predicted T1, a list of time, a list of population and a list of population on the fitted curve
    """

    # create job
    qJob = pulseModel.createQHamiltonian().createJob()
    qJob.setLO(driveX, 0, freq=pulseFreq)
    qJob.addWave(driveX, 0, gaussian(piLen, piAmp, piLen / 2, piLen / 8), t0=0.)
    qJob.addWave(driveX, 0, square(maxIdle, 0), t0=0.)
    stateInit = basis(3, 1)

    # run simulation
    results = pulseModel.ham.simulate(job=qJob, state0=stateInit, isOpen=True, recordEvolution=True)
    states = results[0]['evolution_history']

    # calculate populations from results
    popList = []
    for state in states:
        prj = basis(3, 1) @ dagger(basis(3, 1))
        popList.append(expect(prj, state))
    popList = popList[int(piLen / dt)::]
    tList = list(linspace(0, maxIdle, len(popList)))

    # fit curve
    T1 = fitT1(tList, popList, initFit)
    return T1[0], tList, popList, np.exp(-1. / T1[0] * np.array(tList))


def fitT1(tList: list, popList: list, init: list) -> list:
    r"""
    T1 fitting function, this function takes a list of times, a list of populations and an initial guess for T1.
    The fitting function takes form: :math:`y = \exp(-x / T_1 )`.

    :param tList: a list of times
    :param popList: a list of populations
    :param init: an initial guess
    :result: predicted T1
    """

    # Define fitting function
    def fit(x, t1):
        y = np.exp(-1. / t1 * np.array(x))
        return y

    # Fit curve
    paraFit, _ = curve_fit(fit, tList, popList, init)
    return list(paraFit)


def ramsey(pulseModel: PulseModel, pulseFreq: float, tg: float, x90: float, sample: int = 100,
           maxTime: int = 800, detuning: float = None) -> [list, list]:
    """
    Perform a Ramsey experiment. This function takes a PulseModel object, pulse frequency, pi/2 pulse length,
    pi/2 pulse amplitude, sample size, maximum idling time and detuning pulse amplitude. This function returns
    a list of idling time and a list of population.

    :param pulseModel: a PulseModel object
    :param pulseFreq: pulse frequency
    :param tg: pi/2 pulse length
    :param x90: pi/2 pulse amplitude
    :param sample: sample size
    :param maxTime: maximum idling time
    :param detuning: detuning amplitude
    """

    # get Hamiltonian from pulse model
    ham = pulseModel.createQHamiltonian(frameMode='lab')

    # Define the detuning
    if detuning is None:
        detuning = 2 * pi * 8. / maxTime

    # time list
    tList = np.linspace(0, maxTime, sample)

    # define jobList for Ramsey experiment
    ramseyJob = ham.createJobList()
    for t in tList:
        job = pulseModel.ham.createJob()
        job.setLO(driveX(3), 0, freq=pulseFreq)
        job.addWave(driveX(3), 0, gaussian(t=tg, a=x90 / 2, tau=tg / 2, sigma=tg / 8), t0=0.)  # pi/2 pulse
        job.addWave(driveZ(3), 0, waves=square(t=t, a=detuning), t0=tg)  # simulate the rotation due to the detuning
        job.addWave(driveX(3), 0, gaussian(t=tg, a=-x90 / 2, tau=tg / 2, sigma=tg / 8), t0=tg + t)  # pi/2 pulse
        ramseyJob.addJob(job)

    # run simulation
    stateInit = basis(3, 0)
    results = pulseModel.ham.simulate(state0=stateInit, jobList=ramseyJob, isOpen=True)

    # calculate populations from results
    popList = []
    for result in results:
        rho = result['state']
        prj = basis(3, 1) @ dagger(basis(3, 1))
        popList.append(expect(prj, rho))
    return list(tList), popList


def fitRamsey(t1: float, popList: list, tList: list, detuning: float) -> [float, list]:
    r"""
    Find T2 from Ramsey's result. This function takes a estimated T1 value, a list of population, a list
    of idling time and the amplitude of the detuning. The fitting function takes form:
    :math:`y = - 0.5 \cdot \cos(a \cdot x) \exp(-b \cdot x) + 0.5`

    :param t1: estimated T1.
    :param popList: a list of population from the Ramsey experiment
    :param tList: a list of idling time from the Ramsey experiment
    :param detuning: detuning amplitude
    :return: estimated t2, a list of population on the fitted curve
    """

    # define fitting function
    def fitRam(x, a, b):
        return - np.cos(a * x) * np.exp(- b * x) * 0.5 + 0.5

    # fit curve and obtain t2
    paraFit, _ = curve_fit(fitRam, tList, popList, [detuning, 0.])
    t2 = 1 / (paraFit[1] - 1 / (2 * t1))
    return t2, list(- np.cos(paraFit[0] * array(tList)) * np.exp(- paraFit[1] * array(tList)) * 0.5 + 0.5)


def qubitSpec(pulseModel: PulseModel, freqRange: list, sample: int, amp: float, t: float) -> [list, list]:
    """
    Qubit Spectroscopy. This function finds the qubit frequency by scanning the pulse frequency
    from a user-defined range.

    :param pulseModel: a pulseModel type object
    :param freqRange: a list of LO frequency's range
    :param sample: how many samples to scan within the freqRange
    :param amp: pulse amplitude (preferably Pi-pulse amp)
    :param t: pulse length (preferably Pi-pulse length)
    :return: a list of pulse frequency scanned and a list of population
    """

    # define pulse frequency list
    freqList = linspace(freqRange[0], freqRange[1], sample)

    # initialize job list
    QSpecJob = pulseModel.ham.createJobList()
    for freq in freqList:
        job = pulseModel.ham.createJob()
        job.setLO(driveX(3), 0, freq=freq)
        job.appendWave(driveX, 0, gaussian(t=t, a=amp, tau=t / 2, sigma=t / 8))
        QSpecJob.addJob(job)

    # run simulation and obtain populations from results
    stateInit = basis(3, 0)
    popList = []
    results = pulseModel.ham.simulate(jobList=QSpecJob, state0=stateInit)
    for result in results:
        rho = result['state']
        prj = basis(3, 1) @ dagger(basis(3, 1))
        popList.append(expect(prj, rho))
    return list(freqList), popList



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Default Pulse Generator for QuanlseSchedulerSuperconduct.
"""

from Quanlse.Scheduler import SchedulerPulseGenerator
from Quanlse.QOperation import CircuitLine, Error
from Quanlse.QWaveform import QJob

from Quanlse.remoteOptimizer import remoteOptimize1Qubit as opt1q
from Quanlse.remoteOptimizer import remoteOptimizeCr as optCr
from Quanlse.remoteOptimizer import remoteOptimizeCz as optCz
from Quanlse.remoteOptimizer import remoteOptimizeISWAP as optISWAP


def generate1Q(ham: 'QHamiltonian' = None, cirLine: CircuitLine = None,
               scheduler: 'SchedulerSuperconduct' = None) -> QJob:
    """
    Default generator for single qubit gates.

    :param ham: QHam object containing the system information
    :param cirLine: a CircuitLine object containing the gate information
    :param scheduler: the instance of Quanlse Scheduler Superconducting
    :return: returned QJob object
    """
    subHam = ham.subSystem(cirLine.qRegIndexList)
    if cirLine.data.name in ['X', 'RX']:
        job, inf = opt1q(subHam, cirLine.data.getMatrix(), depth=2, targetInfid=0.0001)
    elif cirLine.data.name in ['Y', 'RY']:
        job, inf = opt1q(subHam, cirLine.data.getMatrix(), depth=2, targetInfid=0.0001)
    elif cirLine.data.name in ['Z', 'S', 'T', 'RZ']:
        job, inf = opt1q(subHam, cirLine.data.getMatrix(), depth=4, targetInfid=0.0001)
    elif cirLine.data.name in ['U', "H"]:
        job, inf = opt1q(subHam, cirLine.data.getMatrix(), depth=4, targetInfid=0.0001)
    else:
        raise Error.ArgumentError(f"Unsupported single qubit gate {cirLine.data.name}.")

    print(f"Infidelity of {cirLine.data.name} on qubit {cirLine.qRegIndexList}: {inf}")

    subHam.job = job
    return subHam.outputInverseJob(ham.subSysNum, ham.sysLevel, ham.dt)


def generateCr(ham: 'QHamiltonian' = None, cirLine: CircuitLine = None,
               scheduler: 'SchedulerSuperconduct' = None) -> QJob:
    """
    Default generator for Cross-resonance gate.

    :param ham: QHam object containing the system information
    :param cirLine: a CircuitLine object containing the gate information
    :param scheduler: the instance of Quanlse Scheduler Superconducting
    :return: returned QJob object
    """
    subHam = ham.subSystem(cirLine.qRegIndexList)
    job, inf = optCr(subHam, (-3.0, 3.0), maxIter=1)
    print(f"Infidelity of {cirLine.data.name} on qubit {cirLine.qRegIndexList}: {inf}")
    subHam.job = job
    return subHam.outputInverseJob(ham.subSysNum, ham.sysLevel, ham.dt)


def generateCz(ham: 'QHamiltonian' = None, cirLine: CircuitLine = None,
               scheduler: 'SchedulerSuperconduct' = None) -> QJob:
    """
    Default generator for controlled-Z gate.

    :param ham: QHam object containing the system information
    :param cirLine: a CircuitLine object containing the gate information
    :param scheduler: the instance of Quanlse Scheduler Superconducting
    :return: returned QJob object
    """
    subHam = ham.subSystem(cirLine.qRegIndexList)
    job, inf = optCz(subHam, tg=40, targetInfidelity=0.01)
    print(f"Infidelity of {cirLine.data.name} on qubit {cirLine.qRegIndexList}: {inf}")
    subHam.job = job
    return subHam.outputInverseJob(ham.subSysNum, ham.sysLevel, ham.dt)


def generateISWAP(ham: 'QHamiltonian' = None, cirLine: CircuitLine = None,
                  scheduler: 'SchedulerSuperconduct' = None) -> QJob:
    """
    Default generator for ISWAP gate.

    :param ham: QHam object containing the system information
    :param cirLine: a CircuitLine object containing the gate information
    :param scheduler: the instance of Quanlse Scheduler Superconducting
    :return: returned QJob object
    """
    subHam = ham.subSystem(cirLine.qRegIndexList)
    job, inf = optISWAP(subHam, tg=50, targetInfidelity=0.01)
    print(f"Infidelity of {cirLine.data.name} on qubit {cirLine.qRegIndexList}: {inf}")
    subHam.job = job
    return subHam.outputInverseJob(ham.subSysNum, ham.sysLevel, ham.dt)


def generatorCloud(ham: 'QHamiltonian') -> SchedulerPulseGenerator:
    """
    Return a default pulse SchedulerPulseGenerator instance for the scheduler.

    :param ham: a Hamiltonian object
    :return: returned generator object
    """

    generator = SchedulerPulseGenerator(ham)

    # Add the generator for single qubit gates
    gateList1q = ['X', 'Y', 'Z', 'H', 'S', 'T', 'RX', 'RY', 'RZ', 'W', 'SQRTW', 'U']
    generator.addGenerator(gateList1q, generate1Q)

    # Add the generator for two-qubit gates
    generator.addGenerator(['CR'], generateCr)
    generator.addGenerator(['CZ'], generateCz)
    generator.addGenerator(['ISWAP'], generateISWAP)

    return generator



#!/usr/bin/python3
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Default Pulse Generator for QuanlseIon.
"""

from math import pi
from numpy import arccos

from Quanlse.QOperation import CircuitLine, Error
from Quanlse.QWaveform import QJob
from Quanlse.QWaveform import QWaveform
from Quanlse.QOperator import QOperator
from Quanlse.Scheduler.SchedulerPulseGenerator import SchedulerPulseGenerator
from Quanlse.remoteOptimizer import remoteIonOptimize1Qubit as opt1q
from Quanlse.remoteOptimizer import remoteIonGeneralMS as optGMS


uWave = QOperator('qubit', None)


def generateIon1Qubit(cirLine: CircuitLine = None, scheduler: 'SchedulerIon' = None) -> QJob:
    """
    Default generator for ion single-qubit gates.

    :param cirLine: input circuit line object
    :param scheduler: input scheduler object
    :return: a returned QJob object
    """
    job = QJob(subSysNum=scheduler.subSysNum, sysLevel=2, dt=scheduler.dt)
    qLabel = cirLine.qRegIndexList
    matrixNumber = cirLine.data.getMatrix()
    theta = arccos(matrixNumber[0][0])
    theta = theta.real
    if cirLine.data.name == 'X':
        amp, inf, ureal = opt1q("ionRx", pi, scheduler.dt)
        print(f"Single gate infidelity on qubit {qLabel[0]} is: \n {inf}")
        job.addWave(uWave, onSubSys=qLabel[0], waves=QWaveform(t0=0.0, t=2 * scheduler.dt, seq=[amp, 0]))
    elif cirLine.data.name == 'Y':
        amp, inf, ureal = opt1q("ionRy", pi, scheduler.dt)
        print(f"Single gate infidelity on qubit {qLabel[0]} is: \n {inf}")
        job.addWave(uWave, onSubSys=qLabel[0], waves=QWaveform(t0=0.0, t=2 * scheduler.dt, seq=[amp, 0]))
    elif cirLine.data.name == 'RX':
        amp, inf, ureal = opt1q("ionRx", theta, scheduler.dt)
        print(f"Single gate infidelity on qubit {qLabel[0]} is: \n {inf}")
        job.addWave(uWave, onSubSys=qLabel[0], waves=QWaveform(t0=0.0, t=2 * scheduler.dt, seq=[amp, 0]))
    elif cirLine.data.name == 'RY':
        amp, inf, ureal = opt1q("ionRy", theta, scheduler.dt)
        print(f"Single gate infidelity on qubit {qLabel[0]} is: \n {inf}")
        job.addWave(uWave, onSubSys=qLabel[0], waves=QWaveform(t0=0.0, t=2 * scheduler.dt, seq=[amp, 0]))
    else:
        raise Error.ArgumentError(f"Unsupported single-qubit gate {cirLine.data.name}.")

    return job


def generateMS(cirLine: CircuitLine = None, scheduler: 'SchedulerIon' = None) -> QJob:
    """
    Default generator for M-S gate.

    :param cirLine: input circuit line object
    :param scheduler: input scheduler object
    :return: a returned QJob object
    """
    if 5 < scheduler.dt < 15:
        job = QJob(subSysNum=scheduler.subSysNum, sysLevel=2, dt=scheduler.dt)
        qLabel = cirLine.qRegIndexList
        gatePair = [[qLabel[0], qLabel[1]]]

        """
        For parameter args1, scheduler.subSysNum is the total ion in chain.
        And simply, we fix the trapped potential with some typical value:
        omegaXY=2 * pi * 2e6, omegaZ=2 * pi * 0.2e6. 
        We choose typical atomMass=171 for Ytterbium atom, 
        and using "transverse" phonon to transmit information.
        """
        args1 = (scheduler.subSysNum, 171, 2 * pi * 2e6, 2 * pi * 0.2e6, "transverse")

        """
        And for quantum gate parameter, we fix the square pulse number
        to be 3*N, where N=scheduler.subSysNum is the ion number in chain.
        The laser detuning mu usually has relation with gate time 
        mu=2*pi/tg, where tg=scheduler.subSysNum * 3 * scheduler.dt.
        """
        args2 = (scheduler.subSysNum * 3, scheduler.dt, 2 * pi / (scheduler.subSysNum * 3 * scheduler.dt))

        result, unitary = optGMS(gatePair, args1=args1, args2=args2)
        infidelity = result['infidelity']
        print(f"Gate infidelity of MS{gatePair[0]} is: \n {infidelity}")
        sequence = result['pulse_list'][0]

        job.addWave(uWave, onSubSys=qLabel[0], waves=QWaveform(t0=0.0, seq=sequence))
        job.addWave(uWave, onSubSys=qLabel[1], waves=QWaveform(t0=0.0, seq=sequence))
    else:
        raise Error.ArgumentError("The sample time of MS gate is between 5 - 15.")

    return job


def defaultPulseGenerator() -> SchedulerPulseGenerator:
    """
    Default pulse generator for ion-trap qubit

    :return: returned generator object
    """
    generator = SchedulerPulseGenerator()

    # Add the generator for single qubit gates
    gateList1q = ['X', 'Y', 'RX', 'RY']
    generator.addGenerator(gateList1q, generateIon1Qubit)

    # Add the generator for MS gates
    generator.addGenerator(['MS'], generateMS)

    return generator




#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Numba Accelerator
"""

from numba import jit
from numpy import ndarray, dot, diag, linalg, exp


@jit(cache=True, nopython=True, parallel=True)
def expm(mat: ndarray) -> ndarray:
    """
    Calculate the matrix exponential :math:`e^{A}` of the input matrix :math:`A`.

    :param mat: the input matrix.
    :return: the matrix exponential.
    """
    # Solving the eigen
    evals, evecs = linalg.eig(mat)
    expEvals = exp(evals)
    d2 = diag(expEvals)
    #  Pardon this godawful circumlocution.
    b = dot(evecs, d2)
    bt = b.transpose()
    at = evecs.transpose()
    et, residuals, rank, s = linalg.lstsq(at, bt)
    e = et.transpose()
    return e



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quanlse Scheduler for Ion Trap quantum computing platform
"""

from Quanlse.Scheduler import Scheduler
from Quanlse.Scheduler.SchedulerPulseGenerator import SchedulerPulseGenerator
from Quanlse.Scheduler.Ion.DefaultPulseGenerator import defaultPulseGenerator


class SchedulerIon(Scheduler):
    """
    Basic class of Quanlse Scheduler for trapped ion platform

    :param dt: AWG sampling time
    :param pulseGenerator: the pulseGenerator object.
    :param subSysNum: size of the subsystem
    :param sysLevel: the energy levels of the system (support different levels for different qubits)
    """
    def __init__(self, dt: float = None, pulseGenerator: SchedulerPulseGenerator = None,
                 subSysNum: int = None, sysLevel: int = None):
        """
        Constructor for class SchedulerIon
        """

        # Initialization
        super().__init__(dt=dt, pulseGenerator=pulseGenerator, subSysNum=subSysNum, sysLevel=sysLevel)

        if self._pulseGenerator is None:
            self._pulseGenerator = defaultPulseGenerator()



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pulse Generator for PulseModel.
"""

from numpy import pi, cos, sin

from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.QHamiltonian import QHamiltonian
from Quanlse.QOperator import uWave, flux, driveX, driveY, driveZ
from Quanlse.QWaveform import QJob, square, gaussian, dragY1, quasiSquareErf, virtualZ, mix
from Quanlse.Scheduler.Superconduct import SchedulerPulseGenerator
from Quanlse.QOperation.RotationGate import RX, RY, RZ
from Quanlse.QOperation.FixedGate import H, CZ
from Quanlse.QOperation import CircuitLine


def pulseGenerator(ham: QHamiltonian) -> SchedulerPulseGenerator:
    r"""
    The pulseGenerator for the simulator.

    :param ham: a QHamiltonian object.
    :return: a SchedulerPulseGenerator object.
    """
    generator = SchedulerPulseGenerator(ham)

    # Add the basic generator for single-qubit gates
    def generateBasic1Q(ham, gate, onQubit, scheduler) -> QJob:
        """ Generate the single-qubit gates """
        job = ham.createJob()

        # Generate the operator
        if gate.name in ['X', 'RX', 'Y', 'RY']:
            if "caliDataXY" not in scheduler.conf.keys():
                raise ArgumentError(f"'caliDataXY' does not exist in scheduler.conf, "
                                    f"hence X/Y controls are not supported.")
            if onQubit not in scheduler.conf["caliDataXY"].keys():
                raise ArgumentError(f"Configuration of qubit(s) {onQubit} does not, "
                                    f"hence X/Y controls are not supported.")
            # Pulse Parameters
            piLen = scheduler.conf["caliDataXY"][onQubit]["piLen"]
            piAmp = scheduler.conf["caliDataXY"][onQubit]["piAmp"]
            piTau = scheduler.conf["caliDataXY"][onQubit]["piTau"]
            piSigma = scheduler.conf["caliDataXY"][onQubit]["piSigma"]
            dragCoef = scheduler.conf["caliDataXY"][onQubit]["dragCoef"]
            if gate.name in ['X', 'RX']:
                amp = piAmp if gate.name == 'X' else gate.uGateArgumentList[0] / pi * piAmp
                piShift = 0.
            elif gate.name in ['Y', 'RY']:
                amp = piAmp if gate.name == 'Y' else gate.uGateArgumentList[0] / pi * piAmp
                piShift = pi / 2
            else:
                raise ArgumentError(f"Unsupported gate {gate.name}!")

            # Add the waveform to job
            _frameMode = scheduler.conf["frameMode"]
            if _frameMode == 'rot':
                job.addWave(driveX, onQubit, gaussian(piLen, amp * cos(piShift), piTau, piSigma), t0=0.)
                job.addWave(driveX, onQubit, dragY1(piLen, dragCoef * amp * sin(- piShift), piTau, piSigma), t0=0.)
                job.addWave(driveY, onQubit, gaussian(piLen, amp * sin(piShift), piTau, piSigma), t0=0.)
                job.addWave(driveY, onQubit, dragY1(piLen, dragCoef * amp * cos(piShift), piTau, piSigma), t0=0.)
            elif _frameMode == 'lab':
                gaussianWave = gaussian(piLen, amp, piTau, piSigma, phase0=piShift)
                dragWave = dragY1(piLen, dragCoef * amp, piTau, piSigma, phase0=piShift)
                job.appendWave(uWave, onQubit, mix(gaussianWave, dragWave))
            else:
                raise ArgumentError(f"Unsupported frameMode '{_frameMode}'!")
        elif gate.name in ['Z', 'RZ']:
            if "caliDataZ" not in scheduler.conf.keys():
                raise ArgumentError(f"'caliDataZ' does not exist in scheduler.conf, "
                                    f"hence Z controls are not supported.")
            if onQubit not in scheduler.conf["caliDataZ"].keys():
                raise ArgumentError(f"Configuration of qubit(s) {onQubit} does not, "
                                    f"hence Z controls are not supported.")
            # Pulse Parameters
            piLen = scheduler.conf["caliDataZ"][onQubit]["piLen"]
            piAmp = scheduler.conf["caliDataZ"][onQubit]["piAmp"]
            amp = piAmp if gate.name == 'Z' else gate.uGateArgumentList[2] / pi * piAmp

            # Add the waveform to job
            job.appendWave(flux, onQubit, square(piLen, amp))

        else:
            raise ArgumentError(f"Unsupported gate {gate.name}!")

        return job

    # Add the generator for single-qubit gates
    def generate1Q(ham, cirLine, scheduler) -> QJob:
        onQubit = int(cirLine.qRegIndexList[0])
        if cirLine.data.name in ['X', 'RX', 'Y', 'RY', 'Z', 'RZ']:
            return generateBasic1Q(ham, cirLine.data, onQubit, scheduler)
        elif cirLine.data.name == 'H':
            jobX = generateBasic1Q(ham, RX(pi), onQubit, scheduler)
            jobY = generateBasic1Q(ham, RY(-pi / 2), onQubit, scheduler)
            return jobX + jobY
        else:
            raise ArgumentError(f"Unsupported gate {cirLine.data.name}!")

    # Add the generator for two-qubit gates
    def generateCz(ham, cirLine, scheduler) -> QJob:
        """ Generate the CZ gate """
        job = ham.createJob()
        qIndex = list(cirLine.qRegIndexList)
        if tuple(qIndex) not in scheduler.conf["caliDataCZ"].keys():
            qIndex.sort()
        qIndex = tuple(qIndex)
        # Pulse parameters
        _frameMode = scheduler.conf["frameMode"]
        czLen = scheduler.conf["caliDataCZ"][qIndex]["czLen"]
        q0ZAmp = scheduler.conf["caliDataCZ"][qIndex]["q0ZAmp"]
        q1ZAmp = scheduler.conf["caliDataCZ"][qIndex]["q1ZAmp"]
        q0VZPhase = scheduler.conf["caliDataCZ"][qIndex]["q0VZPhase"]
        q1VZPhase = scheduler.conf["caliDataCZ"][qIndex]["q1VZPhase"]
        q0ZWave = quasiSquareErf(czLen, q0ZAmp, 0.1 * czLen, 0.9 * czLen, 0.25 * q0ZAmp)
        q1ZWave = quasiSquareErf(czLen, q1ZAmp, 0.1 * czLen, 0.9 * czLen, 0.25 * q1ZAmp)
        if _frameMode == 'lab':
            # Add the square wave
            job.addWave(flux, qIndex[0], q0ZWave, t0=0.)
            job.addWave(flux, qIndex[1], q1ZWave, t0=0.)
            # Add the virtual Z gate phase
            job.addWave(uWave, qIndex[0], virtualZ(q0VZPhase), t0=0.)
            job.addWave(uWave, qIndex[1], virtualZ(q1VZPhase), t0=0.)
        elif _frameMode == 'rot':
            # Add the square wave
            job.addWave(driveZ, qIndex[0], q0ZWave, t0=0.)
            job.addWave(driveZ, qIndex[1], q1ZWave, t0=0.)
            # Add the virtual Z gate phase
            job.addWave(driveX, qIndex[0], virtualZ(q0VZPhase), t0=0.)
            job.addWave(driveY, qIndex[0], virtualZ(q0VZPhase), t0=0.)
            job.addWave(driveX, qIndex[1], virtualZ(q1VZPhase), t0=0.)
            job.addWave(driveY, qIndex[1], virtualZ(q1VZPhase), t0=0.)
        else:
            raise ArgumentError(f"Unsupported frameMode '{_frameMode}'!")
        return job

    def generateCNOT(ham, cirLine, scheduler) -> QJob:
        """ Generate the CNOT gate """
        qIndex = list(cirLine.qRegIndexList)
        jobH1 = generate1Q(ham, CircuitLine(H, [qIndex[1]]), scheduler)
        jobCZ = generateCz(ham, CircuitLine(CZ, qIndex), scheduler)
        return jobH1 + jobCZ + jobH1

    # Add the Generator methods to the generator instance
    generator.addGenerator(['X', 'Y', 'Z', 'RX', 'RY', 'RZ', 'H', 'U'], generate1Q)
    generator.addGenerator(['CZ'], generateCz)
    generator.addGenerator(['CNOT'], generateCNOT)

    return generator



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quantum Operation
"""
from typing import List, Union, Optional, Callable, TYPE_CHECKING

import numpy

from Quanlse.QPlatform import Error

if TYPE_CHECKING:
    from Quanlse.QOperation.FixedGate import FixedGateOP
    from Quanlse.QOperation.RotationGate import RotationGateOP
    from Quanlse.QRegPool import QRegStorage

OperationFunc = Callable[[*'QRegStorage'], None]
RotationArgument = float


class QOperation:
    """
    Basic classes for quantum operation
    """

    def __init__(self, name: Optional[str] = None, bits: Optional[int] = None,
                 matrix: Optional[numpy.ndarray] = None) -> None:
        """
        Constructor for QOperation class
        """
        self.name = name
        self.bits = bits
        self._matrix = matrix

    def getMatrix(self) -> numpy.ndarray:
        """
        Returns a numpy ndarray

        :return: returned matrix in ndarray
        """
        if self.__class__.__name__ == 'FixedGateOP':
            return self._matrix
        elif self.__class__.__name__ == 'RotationGateOP':
            if self._matrix is None:
                self._matrix = self.generateMatrix()
            return self._matrix
        elif self.__class__.__name__ == 'CustomizedGateOP':
            return self._matrix
        else:
            raise Error.ArgumentError(f'{self.__class__.__name__} do not have matrix!')

    def _op(self, qRegList: List['QRegStorage']) -> None:
        """
        Quantum operation base

        :param qRegList: quantum register list
        :return: None
        """
        env = qRegList[0].env
        for qReg in qRegList:
            if qReg.env != env:
                raise Error.ArgumentError('QReg must belong to the same env!')

        if env.__class__.__name__ == 'QProcedure':
            raise Error.ArgumentError('QProcedure should not be operated!')

        if self.bits is not None and self.bits != len(
                qRegList):  # Barrier and QProcedure does not match bits configuration
            raise Error.ArgumentError('The number of QReg must match the setting!')

        if len(qRegList) <= 0:
            raise Error.ArgumentError('Must have QReg in operation!')

        if len(qRegList) != len(set(qReg for qReg in qRegList)):
            raise Error.ArgumentError('QReg of operators in circuit are not repeatable!')

        circuitLine = CircuitLine()
        circuitLine.data = self
        circuitLine.qRegIndexList = [qReg.index for qReg in qRegList]
        env.circuit.append(circuitLine)


Operation = Union[
    'FixedGateOP', 'RotationGateOP']


class CircuitLine:
    """
    This class defines a quantum gate in the quantum circuit model.
    It specifies two key components to characterize a quantum gate:
    The gate and its operated qubit indices.
    """
    data: None  # type: Operation
    qRegIndexList: None  # type: List[int]

    def __init__(self, data: Operation = None, qRegIndexList: List[int] = None):
        r"""
        Initialize a quantum gate instance.

        :param data: a Quanlse.QOperation.Operation instance,
                    the quantum gate to be applied.
        :param qRegIndexList: a list of qubit indices.
                    If `gate` is a single-qubit
                    gate, then `qubits` still be a List of the form `[i]`
        """
        self.data = data
        self.qRegIndexList = qRegIndexList



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simulator Template: 3-qubit pulse simulator.
"""

from numpy import pi
from typing import Dict, Tuple, Union

from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.Simulator import PulseModel
from Quanlse.Scheduler.Superconduct.GeneratorPulseModel import pulseGenerator


def pulseSim3Q(dt: float = 0.5, frameMode: str = 'rot', qubitFreq: Dict[int, float] = None,
               qubitAnharm: Dict[int, float] = None, couplingMap: Dict[Tuple, Union[int, float]] = None) -> PulseModel:
    r"""
    Return a template of 3-qubit simulator.

    :param dt: a sampling time period.
    :param frameMode: indicates the frame, ``rot`` indicates the rotating frame,
        ``lab`` indicates the lab frame.
    :param qubitFreq: the qubit frequency, simulator will use the preset values if None.
    :param qubitAnharm: the qubit anharmonicity, simulator will use the preset values if None.
    :param couplingMap: the coupling between the qubits, simulator will use the preset values if None.
    :return: a 3-qubit PulseModel object.
    """

    # Define parameters needed
    sysLevel = 3  # The system level
    qubitNum = 3  # The number of qubits

    # Coupling map
    if couplingMap is None:
        couplingMap = {
            (0, 1): 0.020 * (2 * pi),
            (1, 2): 0.020 * (2 * pi)
        }
    else:
        pass

    # Qubits frequency anharmonicity
    if qubitAnharm is None:
        qubitAnharm = {
            0: - 0.22 * (2 * pi),
            1: - 0.22 * (2 * pi),
            2: - 0.22 * (2 * pi)
        }
    else:
        if len(qubitAnharm) != qubitNum:
            raise ArgumentError(f"The length of qubit anharmonicity should be {qubitNum}!")

    # Qubit Frequency
    if qubitFreq is None:
        qubitFreq = {
            0: 5.5004 * (2 * pi),
            1: 4.4546 * (2 * pi),
            2: 5.5004 * (2 * pi),
        }
    else:
        if len(qubitFreq) != qubitNum:
            raise ArgumentError(f"The length of qubit frequency should be {qubitNum}!")

    # Drive Frequency
    if frameMode == 'lab':
        # Set the local oscillator
        driveFreq = qubitFreq
    elif frameMode == 'rot':
        driveFreq = None
        pass
    else:
        raise ArgumentError("Only rotating and lab frames are supported!")

    # Generate the pulse model
    model = PulseModel(subSysNum=qubitNum, sysLevel=sysLevel, couplingMap=couplingMap, qubitFreq=qubitFreq,
                       driveFreq=driveFreq, dt=dt, qubitAnharm=qubitAnharm, pulseGenerator=pulseGenerator,
                       frameMode=frameMode)
    model.savePulse = False
    model.conf["frameMode"] = frameMode

    # Single-Qubit gate Calibration data
    model.conf["caliDataXY"] = {
        0: {"piAmp": 0.642614, "piLen": 16.0, "piTau": 8.0, "piSigma": 2.0, "dragCoef": -0.36257},
        1: {"piAmp": 0.637912, "piLen": 16.0, "piTau": 8.0, "piSigma": 2.0, "dragCoef": -0.33307},
        2: {"piAmp": 0.642614, "piLen": 16.0, "piTau": 8.0, "piSigma": 2.0, "dragCoef": -0.36257}
    }

    # Flux pi pulse calibration data
    model.conf["caliDataZ"] = {
        0: {"piAmp": pi / 16., "piLen": 16.0},
        1: {"piAmp": pi / 16., "piLen": 16.0},
        2: {"piAmp": pi / 16., "piLen": 16.0}
    }

    # Two-Qubit gate Calibration data
    model.conf["caliDataCZ"] = {
        (0, 1): {"czLen": 35.00, "q0ZAmp": -3.5570, "q1ZAmp": 1.5022,
                 "q0VZPhase": -1.6924, "q1VZPhase": 1.0337},
        (1, 2): {"czLen": 47.24, "q0ZAmp": 2.0581, "q1ZAmp": -3.5250,
                 "q0VZPhase": 4.5778, "q1VZPhase": -3.9532},
    }

    return model



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Setup Installation
"""

from __future__ import absolute_import
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

setup(
    name='Quanlse',
    version='2.1.0',
    description='A cloud-based platform for quantum control.',
    author='Baidu Quantum',
    author_email='quantum@baidu.com',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'requests',
        'bce-python-sdk-reborn'
    ],
    python_requires='>=3.7, <4',
    license='Apache 2.0'
)